#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.sftp import BaseSftp
from cliboa.util.constant import StepStatus
from cliboa.util.sftp import Sftp


class SftpBaseLoad(BaseSftp):
    def __init__(self):
        super().__init__()


class SftpUpload(SftpBaseLoad):
    """
    Upload file to sftp server
    """

    def __init__(self):
        super().__init__()
        self._quit = False
        self._ignore_empty_file = False

    def quit(self, quit):
        self._quit = quit

    def ignore_empty_file(self, ignore_empty_file):
        self._ignore_empty_file = ignore_empty_file

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._host, self._user, self._src_dir, self._src_pattern, self._dest_dir],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        adaptor = super().get_adaptor()
        if len(files) > 0:
            for file in files:
                if self._ignore_empty_file and os.path.getsize(file) == 0:
                    self._logger.info("0 byte file will no be uploaded %s." % file)
                    continue

                obj = Sftp().put_file(
                    src=file,
                    dest=os.path.join(self._dest_dir, os.path.basename(file)),
                    endfile_suffix=self._endfile_suffix,
                )
                adaptor.execute(obj)
                self._logger.info("%s is successfully uploaded." % file)
        else:
            self._logger.info(
                "Files to upload do not exist. File pattern: {}".format(
                    os.path.join(self._src_dir, self._src_pattern)
                )
            )
            if self._quit is True:
                return StepStatus.SUCCESSFUL_TERMINATION



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
global _STEP_ARGUMENT_CACHE
global _PROCESS_STORE_CACHE
_STEP_ARGUMENT_CACHE = {}
_PROCESS_STORE_CACHE = {}


class StepArgument(object):
    """
    Cache arguments defined in yaml with step name.
    """

    @staticmethod
    def _put(step, instance):
        """
        Store step arguments

        Args:
            step (str): Cache key (step name)
            instance (class): Step class
        """
        props = instance.__dict__
        items = {}
        for k, v in props.items():
            # By default, property keys are changed by python language specification.
            # Adjust key names to the scenario.yaml defined name
            # (both removed underscore and manglinged prefix)
            sp = "_%s_" % instance.__class__.__name__
            if k.startswith(sp):
                k = k.split(sp)[1]
            k = k[1:] if k.startswith("_") else k
            items[k] = v
        _STEP_ARGUMENT_CACHE[step] = items

    @staticmethod
    def get(k):
        """
        Get value

        Args:
            k (str): Cache key

        Returns:
            dict: Value (underscore removed). Returns None if the key does not exist
        """
        return _STEP_ARGUMENT_CACHE.get(k)


class ObjectStore(object):
    """
    Cache any object.
    This cache class is used when same parameter uses from one STEP to the other.
    """

    @staticmethod
    def put(k, v):
        """
        Put value

        Args:
            k (str): Cache key
            v (dict): Cache value
        """
        _PROCESS_STORE_CACHE[k] = v

    @staticmethod
    def get(k):
        """
        Get value

        Args:
            k (str): Cache key

        Returns:
            dict: Value. Returns None if the key does not exist
        """
        return _PROCESS_STORE_CACHE.get(k)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

from cliboa.adapter.sftp import SftpAdapter
from cliboa.scenario.base import BaseStep


class BaseSftp(BaseStep):
    def __init__(self):
        super().__init__()

        self._src_dir = None
        self._src_pattern = None
        self._dest_dir = ""
        self._host = None
        self._port = 22
        self._user = None
        self._password = None
        self._key = None
        self._passphrase = None
        self._endfile_suffix = None
        self._timeout = 30
        self._retry_count = 3

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def host(self, host):
        self._host = host

    def port(self, port):
        self._port = port

    def user(self, user):
        self._user = user

    def password(self, password):
        self._password = password

    def key(self, key):
        self._key = key

    def passphrase(self, passphrase):
        self._passphrase = passphrase

    def endfile_suffix(self, endfile_suffix):
        self._endfile_suffix = endfile_suffix

    def timeout(self, timeout):
        self._timeout = timeout

    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def get_adaptor(self):
        if isinstance(self._key, str):
            self._logger.warning(
                (
                    "DeprecationWarning: "
                    "In the near future, "
                    "the `key` will be changed to accept only dictionary types. "
                    "Please see more information "
                    "https://github.com/BrainPad/cliboa/blob/master/docs/modules/sftp_download.md"
                )
            )
            key_filepath = self._key
        else:
            key_filepath = self._source_path_reader(self._key)

        return SftpAdapter(
            host=self._host,
            user=self._user,
            password=self._password,
            key=key_filepath,
            passphrase=self._passphrase,
            timeout=self._timeout,
            retryTimes=self._retry_count,
            port=self._port,
        )



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.sqlite import SqliteTransaction
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.csv import Csv
from cliboa.util.exception import CliboaException, FileNotFound


class SqliteImport(SqliteTransaction):
    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._tblname = None
        self._primary_key = None
        self._index = []
        self._refresh = True
        self._force_insert = False
        self._encoding = "utf-8"

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def tblname(self, tblname):
        self._tblname = tblname

    def primary_key(self, primary_key):
        self._primary_key = primary_key

    def index(self, index):
        self._index = index

    def refresh(self, refresh):
        self._refresh = refresh

    def force_insert(self, force_insert):
        self._force_insert = force_insert

    def encoding(self, encoding):
        self._encoding = encoding

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._src_pattern, self._tblname]
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self._logger.info("Files found %s" % files)

        if len(files) == 0:
            raise FileNotFound("No csv file was found.")

        files.sort()

        def func():
            # Find csv columns from all csv files
            csv_columns = []
            for file in files:
                csv_columns.extend(Csv.get_column_names(file))
            csv_columns = sorted(set(csv_columns), key=csv_columns.index)

            if self._refresh is True:
                # Drop table in advance, If refresh is True
                self._sqlite_adptr.drop_table(self._tblname)
                self._sqlite_adptr.create_table(self._tblname, csv_columns, self._primary_key)
            else:
                self._sqlite_adptr.create_table(self._tblname, csv_columns, self._primary_key)

                if self._force_insert is True:
                    db_columns = self._sqlite_adptr.get_column_names(self._tblname)
                    result = list(set(csv_columns) - set(db_columns))
                    self._sqlite_adptr.add_columns(self._tblname, result)
                else:
                    # Make sure if csv columns and db table names are exactly the same
                    db_columns = self._sqlite_adptr.get_column_names(self._tblname)
                    if self._sqlite_adptr.escape_columns(
                        csv_columns
                    ) != self._sqlite_adptr.escape_columns(db_columns):
                        raise CliboaException(
                            "Csv columns %s were not matched to table column %s."
                            % (csv_columns, db_columns)
                        )

            for file in files:
                self._sqlite_adptr.import_table(
                    file, self._tblname, refresh=False, encoding=self._encoding
                )

            if self._index and len(self._index) > 0:
                """
                Create index (Add the index at the end for
                better performance when insert data is large)
                """
                self._sqlite_adptr.add_index(self._tblname, self._index)

        super().execute(func)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
from time import sleep

from paramiko import AutoAddPolicy, SSHClient


class SftpAdapter(object):
    """
    Sftp Adaptor
    """

    TIMEOUT_SEC = 30

    def __init__(
        self,
        host,
        user,
        password=None,
        key=None,
        passphrase=None,
        timeout=TIMEOUT_SEC,
        retryTimes=3,
        port=22,
    ):
        """
        Must set whether password or key

        Args:
            host (str): hostname
            user (str): username
            password (str): password
            key (path): private key path
            passphrase (str): passphrase
            timeout=30 (int): timeout seconds
            retryTimes=3 (int): retry count
            port=22 (int): port number
        Raises:
            ValueError: both password and private key are empty, or both specified.
        """

        logging.getLogger("paramiko").setLevel(logging.CRITICAL)

        if (not password and not key) or (password and key):
            raise ValueError("Illegal paramters are given.")

        self._host = host
        self._user = user
        self._password = password
        self._key = key
        self._passphrase = passphrase
        self._timeout = timeout
        self._retryTimes = retryTimes
        self._port = 22 if port is None else port
        self._logger = logging.getLogger(__name__)

    def execute(self, obj):
        """
        Execute sftp process.
        Pass a result of cliboa.util.sftp.Sftp.{{ function }} for the arguments.

        Args:
            obj (tuple):
                obj[0]: A function to be executed.
                obj[1]: Arguments when function is executed.
        """
        return self._execute(obj[0], obj[1])

    def _execute(self, func, kwargs):

        if not func:
            raise ValueError("Function must not be empty.")

        err = None
        for _ in range(self._retryTimes):
            ssh = None
            sftp = None

            try:
                ssh = SSHClient()
                ssh.set_missing_host_key_policy(AutoAddPolicy())
                if self._password:
                    ssh.connect(
                        hostname=self._host,
                        username=self._user,
                        password=self._password,
                        port=self._port,
                        timeout=self._timeout,
                    )
                else:
                    ssh.connect(
                        hostname=self._host,
                        username=self._user,
                        key_filename=self._key,
                        passphrase=self._passphrase,
                        port=self._port,
                        timeout=self._timeout,
                    )

                sftp = ssh.open_sftp()

                # call the argument function
                ret = func(sftp=sftp, **kwargs)

                return ret

            except Exception as e:
                self._logger.warning(e)
                self._logger.warning(kwargs)
                err = e

            finally:
                if sftp is not None:
                    sftp.close()
                if ssh is not None:
                    ssh.close()

            self._logger.warning("Unexpected error occurred. Retry will start in 10 sec.")
            sleep(10)

        raise IOError(err, "SFTP failed.")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

from azure.storage.blob import BlobServiceClient


class BlobServiceAdapter(object):
    def get_client(self, account_url, account_access_key, connection_string):
        if connection_string:
            return BlobServiceClient.from_connection_string(connection_string)
        else:
            return BlobServiceClient(account_url=account_url, credential=account_access_key)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import random
import string

from cliboa.util.lisboa_log import LisboaLog


class StringUtil(object):
    def __init__(self):
        self._logger = LisboaLog.get_logger(__name__)

    def random_str(self, length):
        """
        Generate a random string

        Args:
            length: Length of characters

        Returns:
            str: New randomly generated string
        """
        return "".join([random.choice(string.ascii_letters + string.digits) for i in range(length)])



import logging


class ExceptionHandler(object):
    def __init__(self):
        super().__init__()
        self._logger = logging.getLogger(__name__)

    def handle_error(self, *args, **kwargs):
        # Please implement in a subclass if you would like to do something.
        self._logger.warning(args)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.base import BaseStep


class FileWrite(BaseStep):
    """
    Load file to server
    """

    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._dest_path = None
        self._encoding = "utf-8"
        self._mode = "a"

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_path(self, dest_path):
        self._dest_path = dest_path

    def encoding(self, encoding):
        self._encoding = encoding

    def mode(self, mode):
        self._mode = mode



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
import os
from abc import abstractmethod
from collections import OrderedDict

import yaml

from cliboa.core.validator import (
    EssentialKeys,
    ScenarioJsonKey,
    ScenarioJsonType,
    ScenarioYamlKey,
    ScenarioYamlType,
)
from cliboa.util.lisboa_log import LisboaLog


class ScenarioParser(object):
    """
    Base class of scenario file parser
    """

    def __init__(self, pj_scenario_file, cmn_scenario_file):
        self._logger = LisboaLog.get_logger(__name__)
        self._pj_scenario_file = pj_scenario_file
        self._cmn_scenario_file = cmn_scenario_file

    @abstractmethod
    def parse(self):
        """
        Parse scenario file
        """


class YamlScenarioParser(ScenarioParser):
    """
    scenario.yml parser
    """

    def parse(self):
        self._logger.info("Start to parse scenario.yml")

        # Load projet scenario file
        with open(self._pj_scenario_file, "r") as pj_f:
            pj_yaml_dict = yaml.safe_load(pj_f)
            self._valid_scenario_yaml(pj_yaml_dict)
            self._exists_ess_keys(pj_yaml_dict["scenario"])

        # Load common scenario file (if exist)
        cmn_yaml_dict = None
        if os.path.isfile(self._cmn_scenario_file):
            with open(self._cmn_scenario_file, "r") as cmn_f:
                cmn_yaml_dict = yaml.safe_load(cmn_f)
                self._valid_scenario_yaml(cmn_yaml_dict)
                self._exists_ess_keys(cmn_yaml_dict["scenario"])

        if cmn_yaml_dict:
            yaml_list = self._merge_scenario_yaml(
                pj_yaml_dict["scenario"], cmn_yaml_dict["scenario"]
            )
        else:
            yaml_list = pj_yaml_dict.get("scenario")

        self._logger.info("Finish to parse scenario.yml")
        return yaml_list

    def _valid_scenario_yaml(self, yaml_dict):
        """
        validate instance type and essential key in scenario.yml
        """
        valids = [ScenarioYamlKey, ScenarioYamlType]
        for valid in valids:
            valid(yaml_dict)()

    def _merge_scenario_yaml(self, pj_yaml_list, cmn_yaml_list):
        """
        Merge project scenario.yml and common scenario.yml.
        If the same class specification exists,
        scenario.yml of projet is taken priority.
        """
        for pj_yaml_dict in pj_yaml_list:
            # If same class exists, merge arguments
            if pj_yaml_dict.get("parallel"):
                for row in pj_yaml_dict.get("parallel"):
                    self._merge(row, cmn_yaml_list)
            elif pj_yaml_dict.get("parallel_with_config"):
                steps = pj_yaml_dict.get("parallel_with_config").get("steps")
                for row in steps:
                    self._merge(row, cmn_yaml_list)
            else:
                self._merge(pj_yaml_dict, cmn_yaml_list)

        return pj_yaml_list

    def _merge(self, pj_yaml_dict, cmn_yaml_list):
        cmn_yaml_dict_in_list = [
            d for d in cmn_yaml_list if d.get("class") == pj_yaml_dict.get("class")
        ]
        if not cmn_yaml_dict_in_list:
            return

        pj_cls_attrs = pj_yaml_dict.get("arguments", "")
        cmn_cls_attrs = cmn_yaml_dict_in_list[0].get("arguments")

        # Merge arguments
        if pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = dict(cmn_cls_attrs, **pj_cls_attrs)
        elif not pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = cmn_cls_attrs
        pj_yaml_dict["arguments"] = pj_cls_attrs

    def _exists_ess_keys(self, scenario_yaml_list):
        """
        Check if the essential keys exist in scenario.yml
        """
        valid = EssentialKeys(scenario_yaml_list)
        valid()


class JsonScenarioParser(ScenarioParser):
    """
    scenario.json parser
    """

    def parse(self):
        self._logger.info("Start to parse scenario.json")

        # Load projet scenario file
        with open(self._pj_scenario_file, "r") as pj_f:
            pj_json_dict = json.load(pj_f, object_pairs_hook=OrderedDict)
            self._valid_scenario_json(pj_json_dict)
            self._exists_ess_keys(pj_json_dict["scenario"])

        # Load common scenario file (if exist)
        cmn_json_dict = None
        if os.path.isfile(self._cmn_scenario_file):
            with open(self._cmn_scenario_file, "r") as cmn_f:
                cmn_json_dict = json.load(cmn_f, object_pairs_hook=OrderedDict)
                self._valid_scenario_json(cmn_json_dict)
                self._exists_ess_keys(cmn_json_dict["scenario"])

        if cmn_json_dict:
            json_list = self._merge_scenario_json(
                pj_json_dict["scenario"], cmn_json_dict["scenario"]
            )
        else:
            json_list = pj_json_dict.get("scenario")

        self._logger.info("Finish to parse scenario.json")
        return json_list

    def _valid_scenario_json(self, json_dict):
        """
        validate instance type and essential key in scenario.json
        """
        valids = [ScenarioJsonKey, ScenarioJsonType]
        for valid in valids:
            valid(json_dict)()

    def _merge_scenario_json(self, pj_json_list, cmn_json_list):
        """
        Merge project scenario.json and common scenario.json.
        If the same class specification exists,
        scenario.yml of projet is taken priority.
        """
        for pj_json_dict in pj_json_list:
            # If same class exists, merge arguments
            if pj_json_dict.get("parallel"):
                for row in pj_json_dict.get("parallel"):
                    self._merge(row, cmn_json_list)
            elif pj_json_dict.get("parallel_with_config"):
                steps = pj_json_dict.get("parallel_with_config").get("steps")
                for row in steps:
                    self._merge(row, cmn_json_list)
            else:
                self._merge(pj_json_dict, cmn_json_list)

        return pj_json_list

    def _merge(self, pj_json_dict, cmn_json_list):
        cmn_json_dict_in_list = [
            d for d in cmn_json_list if d.get("class") == pj_json_dict.get("class")
        ]
        if not cmn_json_dict_in_list:
            return

        pj_cls_attrs = pj_json_dict.get("arguments", "")
        cmn_cls_attrs = cmn_json_dict_in_list[0].get("arguments")

        # Merge arguments
        if pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = dict(cmn_cls_attrs, **pj_cls_attrs)
        elif not pj_cls_attrs and cmn_cls_attrs:
            pj_cls_attrs = cmn_cls_attrs
        pj_json_dict["arguments"] = pj_cls_attrs

    def _exists_ess_keys(self, scenario_json_list):
        """
        Check if the essential keys exist in scenario.yml
        """
        valid = EssentialKeys(scenario_json_list)
        valid()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re
import subprocess
from abc import abstractmethod

from cliboa.conf import env
from cliboa.core.file_parser import JsonScenarioParser, YamlScenarioParser
from cliboa.core.listener import StepStatusListener
from cliboa.core.scenario_queue import ScenarioQueue
from cliboa.core.step_queue import StepQueue
from cliboa.core.validator import ProjectDirectoryExistence, ScenarioFileExistence
from cliboa.scenario import *  # noqa
from cliboa.util.cache import StepArgument
from cliboa.util.class_util import ClassUtil
from cliboa.util.exception import InvalidParameter, ScenarioFileInvalid
from cliboa.util.helper import Helper
from cliboa.util.lisboa_log import LisboaLog
from cliboa.util.parallel_with_config import ParallelWithConfig

__all__ = ["YamlScenarioManager", "JsonScenarioManager"]


class ScenarioManager(object):
    """
    Base class which is to create individual instances from scenario files,
    and push them to the executable queue.

    Note : Currently only yaml format scenario file is implemented.
    """

    def __init__(self, cmd_args):
        self._logger = LisboaLog.get_logger(__name__)
        self._cmd_args = cmd_args
        self._pj_dir = os.path.join(env.PROJECT_DIR, cmd_args.project_name)
        self._pj_scenario_dir = os.path.join(
            env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_DIR_NAME
        )
        if cmd_args.format == "yaml":
            self._pj_scenario_file = (
                os.path.join(env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_FILE_NAME)
                + ".yml"
            )
            self._cmn_scenario_file = os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME) + ".yml"
        else:
            self._pj_scenario_file = (
                os.path.join(env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_FILE_NAME)
                + "."
                + cmd_args.format
            )
            self._cmn_scenario_file = (
                os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME) + "." + cmd_args.format
            )
        self._replace_vars_pattern = re.compile(r"{{(.*?)}}")

    def create_scenario_queue(self):
        # validation
        self._valid_essential_dir()
        self._valid_essential_files()

        scenario_list = self.parse_file()
        if not scenario_list or isinstance(scenario_list, list) is False:
            raise ScenarioFileInvalid("scenario file is invalid.")

        self._logger.info("Start to create scenario queue")
        queue = StepQueue()
        self._add_queue(queue, scenario_list)

        # save queue to static area
        setattr(ScenarioQueue, "step_queue", queue)
        self._logger.info("Finish to invoke scenario")

    @abstractmethod
    def parse_file(self, file):
        """
        Parse file contents to the list object
        """

    def _valid_essential_dir(self):
        """
        Project directory validation
        """
        valid_instance = ProjectDirectoryExistence()
        valid_instance(self._pj_dir)

    def _valid_essential_files(self):
        """
        Scenario file validation
        """
        valid_instance = ScenarioFileExistence()
        valid_instance(self._pj_scenario_file)

    def _add_queue(self, queue, scenario_list):
        """
        Add executable instance to the queue
        """
        for block in scenario_list:
            if "multi_process_count" in block.keys():
                Helper.set_property(queue, "multi_proc_cnt", block.get("multi_process_count"))
            elif "force_continue" in block.keys():
                Helper.set_property(queue, "force_continue", block.get("force_continue"))
            else:
                instance = self._create_executable_instances(block)
                queue.push(instance)

        self._logger.info("Finish to create scenario queue")

    def _create_executable_instances(self, s_dict):
        """
        Create executable instances

        Returns:
            list: Executable instances
        """
        instances = []
        if "parallel" in s_dict.keys():
            for row in s_dict.get("parallel"):
                instance = self._create_instance(row)
                instances.append(instance)
                StepArgument._put(row["step"], instance)
        elif "parallel_with_config" in s_dict.keys():
            steps_config_block = s_dict.get("parallel_with_config")
            steps, config = self._split_steps_config(steps_config_block)
            parallel = ParallelWithConfig([], config)
            for row in steps:
                instance = self._create_instance(row)
                parallel.steps.append(instance)
                StepArgument._put(row["step"], instance)
            instances.append(parallel)
        else:
            instance = self._create_instance(s_dict)
            instances.append(instance)
            StepArgument._put(s_dict["step"], instance)

        return instances

    def _create_instance(self, s_dict):
        """
        Create instance
        """
        cls_name = s_dict["class"]
        self._logger.debug("Create %s instance" % cls_name)

        if ClassUtil().is_custom_cls(cls_name) is True:
            from cliboa.core.factory import CustomInstanceFactory

            instance = CustomInstanceFactory.create(cls_name)
        else:
            cls = globals()[cls_name]
            instance = cls()

        cls_attrs_dict = {}
        if "arguments" in s_dict.keys():
            cls_attrs_dict = s_dict["arguments"]

        values = {}
        if cls_attrs_dict:
            cls_attrs_dict, with_vars = self._split_class_vars(cls_attrs_dict)
            ret = self._replace_arguments(cls_attrs_dict, with_vars)
            for dict_k, dict_v in ret.items():
                Helper.set_property(instance, dict_k, dict_v)
            values.update(ret)

        base_args = ["step", "symbol", "parallel", "listeners"]
        for arg in base_args:
            if arg == "listeners":
                self._append_listeners(instance, s_dict.get(arg), values)
            else:
                Helper.set_property(instance, arg, s_dict.get(arg))

        Helper.set_property(
            instance,
            "logger",
            LisboaLog.get_logger(instance.__class__.__name__),
        )

        return instance

    def _split_class_vars(self, arguments):
        """
        If "with_vars" exist in arguments of individual steps,
        split into two(arguments except with_vars and with_vars)

        Args:
            arguments (dict): Arguments of steps

        Returns:
            tuple: (step attribute, with_vars parameter)
        """
        exists_with_vars = "with_vars" in arguments.keys()
        with_vars = {}
        if exists_with_vars:
            variables = arguments["with_vars"]
            for yaml_k, yaml_v in variables.items():
                with_vars[yaml_k] = yaml_v
            del arguments["with_vars"]
        return arguments, with_vars

    def _replace_arguments(self, arguments, with_vars):
        """
        Nested replacement of argments.

        First args:
            arguments (dict): Arguments of steps
            with_vars (dict): with_vars parameter

        Returns:
            dict: Dictionary of arguments which was replaced.
        """
        if isinstance(arguments, dict):
            return {
                dict_k: self._replace_arguments(dict_v, with_vars)
                for dict_k, dict_v in arguments.items()
            }
        elif isinstance(arguments, list):
            return [self._replace_arguments(list_v, with_vars) for list_v in arguments]
        elif isinstance(arguments, str):
            matches = self._replace_vars_pattern.findall(arguments)
            for match in matches:
                var_name = match.strip()
                if not var_name:
                    raise InvalidParameter("Alternative argument was empty.")
                if var_name.startswith("env."):
                    arguments = self._replace_envs(arguments, var_name)
                else:
                    cmd = with_vars[var_name]
                    if not cmd:
                        raise ScenarioFileInvalid(
                            "scenario file is invalid. 'with_vars' definition against %s does not exist."  # noqa
                            % var_name
                        )
                    arguments = self._replace_vars(arguments, var_name, cmd)
            return arguments
        else:
            return arguments

    def _replace_vars(self, yaml_v, var_name, cmd):
        """
        This method replaces the value of {{ xxx }} based on the result of the shell script.

        Ex.
          -- IN --
          yaml_v: /resources/{{ yyyyMMdd }}
          var_name: yyyyMMdd
          cmd: date '+%Y%m%d'

          -- OUT --
          /resources/20210101

        Args:
            yaml_v: Yaml value. Must contain {{ xxx }}
            var_name: Name of xxx
            cmd: Shell script

        Returns:
            str: replaced value
        """
        shell_output = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
        shell_output = shell_output.strip()
        # remove head byte string
        shell_output = re.sub("^b", "", str(shell_output))
        # remove '
        shell_output = re.sub("'", "", str(shell_output))

        return re.sub(r"{{(\s?)%s(\s?)}}" % var_name, shell_output, yaml_v)

    def _replace_envs(self, yaml_v, var_name):
        """
        This method replaces the value of {{ env.xxx }} from system environment values.
        If values are not found in environment, key error will be raised.

        Args:
            yaml_v: Yaml value. Must contain {{ env.xxx }}
            var_name: Name of env.xxx

        Returns:
            str: replaced value
        """
        env_value = os.environ[var_name[4:]]
        return re.sub(r"{{(\s?)%s(\s?)}}" % var_name, env_value, yaml_v)

    def _split_steps_config(self, block):
        """
        If "config" exist in arguments of parallel_with_config,
        split into two(list of steps and config)

        Args:
            arguments (dict): List of steps

        Returns:
            tuple: (steps, config parameter)
        """
        exists_config = "config" in block.keys()
        config = {}
        if exists_config:
            variables = block["config"]
            for k, v in variables.items():
                config[k] = v
        steps = []
        if "steps" in block.keys():
            steps = block["steps"]
        return steps, config

    def _append_listeners(self, instance, args, values):
        listeners = [StepStatusListener()]

        if args is not None:
            from cliboa.core.factory import CustomInstanceFactory

            if type(args) is str:
                clz = CustomInstanceFactory.create(args)
                clz.__dict__.update(values)
                listeners.append(clz)
            elif type(args) is list:
                for arg in args:
                    clz = CustomInstanceFactory.create(arg)
                    clz.__dict__.update(values)
                    listeners.append(CustomInstanceFactory.create(clz))
        Helper.set_property(instance, "listeners", listeners)


class YamlScenarioManager(ScenarioManager):
    """
    Create instances from yaml format scenario file.
    """

    def parse_file(self):
        """
        Parse yaml format file to list object
        """
        parser = YamlScenarioParser(self._pj_scenario_file, self._cmn_scenario_file)
        return parser.parse()


class JsonScenarioManager(ScenarioManager):
    """
    Create instances from json format scenario file.
    """

    def parse_file(self):
        """
        Parse json format file to list object
        """
        parser = JsonScenarioParser(self._pj_scenario_file, self._cmn_scenario_file)
        return parser.parse()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import subprocess

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.base import BaseStep


class ExecuteShellScript(BaseStep):
    """
    Execute Shell Script
    """

    def __init__(self):
        super().__init__()
        self._command = None
        self._work_dir = "./"

    def command(self, command):
        self._command = command

    def work_dir(self, work_dir):
        self._work_dir = work_dir

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._command],
        )
        valid()

        # Setting Up Directory
        default_dir = os.getcwd()

        try:
            if self._work_dir:
                os.chdir(self._work_dir)

            # Run Commands
            content = self._command.get("content")
            file_path = self._command.get("file")

            if content:
                for command in content.split("&&"):
                    subprocess.run(command.strip().split(" "))
            elif file_path:
                subprocess.call(file_path)
        finally:
            # Set Directory to default
            if self._work_dir:
                os.chdir(default_dir)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
import os
import stat


class Sftp(object):
    def list_files(self, dir, dest, pattern, endfile_suffix=None, ignore_empty_file=False):
        """
        Fetch all the files in specified directory

        Args:
            dir (str): fetch target directory
            dest (str): local directory to save files
            pattern (object): fetch file pattern
            endfile_suffix=None (str): Download a file only if "filename + endfile_suffix" is exists
            ignore_empty_file=False (bool): If True, size zero files are not be downloaded

        Returns (tuple):
            func: list_file_func
            params: parameters for list_file_func

        Raises:
            IOError: failed to get data
        """
        return (
            list_file_func,
            {
                "dir": dir,
                "dest": dest,
                "pattern": pattern,
                "endfile_suffix": endfile_suffix,
                "ignore_empty_file": ignore_empty_file,
            },
        )

    def clear_files(self, dir, pattern):
        """
        Remove all the files which matche to pattern. Not remove directory.

        Args:
            dir (str): remove target dir
            pattern (object): remove target file pattern

        Returns (tuple):
            func: clear_file_func
            params: parameters for clear_file_func

        Raises:
            IOError: failed to remove
        """
        return (clear_file_func, {"dir": dir, "pattern": pattern})

    def remove_specific_file(self, dir, fname):
        """
        Remove specific file (by name) from target directory

        Args:
            dir (str): remove target dir
            fname (str): file name (exact match)

        Returns (tuple):
            func: remove_specific_file_func
            params: parameters for remove_specific_file_func

        Raises:
            IOError: failed to remove
        """
        return (remove_specific_file_func, {"dir": dir, "fname": fname})

    def get_specific_file(self, src, dest):
        """
        Fetch file by exact pattern match

        Args:
            src (str): fetch target absolute path
            dest (str): local directory to save file

        Returns (tuple):
            func: get_specific_file_func
            params: parameters for get_specific_file_func

        Raises:
            IOError: failed to get data
        """
        return (get_specific_file_func, {"src": src, "dest": dest})

    def put_file(self, src, dest, endfile_suffix=None):
        """
        Upload file to sftp server

        Args:
            src (str): local file to upload
            dest (str): destination sftp path to upload
            endfile_suffix=None (str): Places file with original file name
                            + "endfile_suffix" when upload completed
        Args:
            src (str): fetch target absolute path
            dest (str): local directory to save file

        Returns (tuple):
            func: put_file_func
            params: parameters for put_file_func

        Raises:
            IOError: failed to upload
        """
        return (put_file_func, {"src": src, "dest": dest, "endfile_suffix": endfile_suffix})

    def file_exists_check(self, dir, pattern, ignore_empty_file=False):
        """
        Fetch all the files name in specified directory

        Args:
            dir (str): fetch target directory
            pattern (object): fetch file pattern
            ignore_empty_file=False (bool):
             If True, It is treated as if there is no file with size 0.

        Returns (tuple):
            func: file_exists_check_func
            params: parameters for file_exists_check_func

        Raises:
            IOError: failed to get data
        """
        return (
            file_exists_check_func,
            {"dir": dir, "pattern": pattern, "ignore_empty_file": ignore_empty_file},
        )


def list_file_func(**kwargs):
    """
    Get all the files which matches to pattern
    """
    endfile_suffix = kwargs["endfile_suffix"]
    targets = kwargs["sftp"].listdir(kwargs["dir"])
    files = []
    for f in targets:
        if kwargs["pattern"].fullmatch(f) is None:
            continue
        if endfile_suffix and not f + endfile_suffix in targets:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            if kwargs["ignore_empty_file"] is True:
                size = _get_file_size(kwargs["sftp"], fpath)
                if size == 0:
                    continue
            kwargs["sftp"].get(fpath, os.path.join(kwargs["dest"], f))
            files.append(f)
    return files


def clear_file_func(**kwargs):
    """
    All the files which match to pattern.
    """
    files = []
    for f in kwargs["sftp"].listdir(kwargs["dir"]):
        if kwargs["pattern"].fullmatch(f) is None:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            kwargs["sftp"].remove(fpath)
            files.append(fpath)
    return files


def remove_specific_file_func(**kwargs):
    """
    Remove specific file (by name) from target directory
    """
    kwargs["sftp"].remove(os.path.join(kwargs["dir"], kwargs["fname"]))


def get_specific_file_func(**kwargs):
    """
    Get only one files which matches to pattern
    """
    kwargs["sftp"].get(kwargs["src"], kwargs["dest"])


def put_file_func(**kwargs):
    """
    Function that is to upload a file.
    '.' as prefix is added to the file name
    while uploading and then the prefix '.'
    will be removed (rename the file) when upload complete.
    """
    dirname = os.path.dirname(kwargs["dest"])

    # Create directory if it doesn't exist
    try:
        kwargs["sftp"].stat(dirname)
    except FileNotFoundError:
        kwargs["sftp"].mkdir(dirname)

    tmp_dest = os.path.join(dirname, "." + os.path.basename(kwargs["dest"]))

    _logger = logging.getLogger(__name__)

    if _logger.isEnabledFor(logging.DEBUG):

        def cb(sent, size):
            _logger.debug("Transfer %s / %s" % (sent, size))

        file_size = os.stat(kwargs["src"]).st_size
        with open(kwargs["src"], "rb") as fl:
            _logger.debug("Open src file")
            with kwargs["sftp"].file(tmp_dest, "wb") as fr:
                _logger.debug("Open dest file")
                fr.set_pipelined(True)
                _transfer_with_callback(reader=fl, writer=fr, file_size=file_size, callback=cb)
            _logger.debug("End")
    else:
        kwargs["sftp"].put(kwargs["src"], tmp_dest)

    # Same file name is removed in advance, if exists
    try:
        f = kwargs["sftp"].stat(kwargs["dest"])
        if f:
            kwargs["sftp"].remove(kwargs["dest"])
    except FileNotFoundError:
        pass

    kwargs["sftp"].rename(tmp_dest, kwargs["dest"])

    endfile_suffix = kwargs["endfile_suffix"]
    if endfile_suffix is not None:
        endfile = kwargs["src"] + endfile_suffix
        open(endfile, mode="w").close()
        kwargs["sftp"].put(endfile, kwargs["dest"] + endfile_suffix)
        os.remove(endfile)


def file_exists_check_func(**kwargs):
    """
    Get all the file names which matches to pattern
    """
    files = []
    for f in kwargs["sftp"].listdir(kwargs["dir"]):
        if kwargs["pattern"].fullmatch(f) is None:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            if kwargs["ignore_empty_file"] is True:
                if _get_file_size(kwargs["sftp"], fpath) == 0:
                    continue
            files.append(f)
    return files


def _transfer_with_callback(reader, writer, file_size, callback):
    """
    For Debug
    """
    size = 0
    _logger = logging.getLogger(__name__)
    _logger.debug("Call back method")
    while True:
        data = reader.read(32768)
        writer.write(data)
        size += len(data)
        if len(data) == 0:
            _logger.debug("No data")
            break
        if callback is not None:
            callback(size, file_size)
    _logger.debug("Size %s" % size)
    return size


def _is_file(sftp, path):
    """
    Returns True if target object is a file, False if not.
    """
    info = sftp.stat(path)
    if stat.S_ISREG(info.st_mode):
        return True
    else:
        return False


def _get_file_size(sftp, path):
    """
    Returns a file size
    """
    info = sftp.stat(path)
    return info.st_size



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv

from cliboa.util.exception import CliboaException


class Csv(object):
    @staticmethod
    def quote_convert(string):
        """
        Convert string to csv quoting type.
        """
        convert_type = {
            "QUOTE_ALL": csv.QUOTE_ALL,
            "QUOTE_MINIMAL": csv.QUOTE_MINIMAL,
            "QUOTE_NONNUMERIC": csv.QUOTE_NONNUMERIC,
            "QUOTE_NONE": csv.QUOTE_NONE,
        }

        if string.upper() not in convert_type:
            raise CliboaException(
                "Unknown string. One of the followings are allowed [QUOTE_ALL, QUOTE_MINIMAL, QUOTE_NONNUMERIC, QUOTE_NONE]"  # noqa
            )
        return convert_type.get(string.upper())

    @staticmethod
    def delimiter_convert(string):
        """
        Convert string to csv delimiter.
        """
        if string.upper() == "CSV":
            return ","
        elif string.upper() == "TSV":
            return "\t"
        else:
            raise CliboaException("Unknown string. One of the followings are allowed [CSV, TSV]")

    @staticmethod
    def newline_convert(string):
        convert_type = {
            "LF": "\n",
            "CR": "\r",
            "CRLF": "\r\n",
        }

        if string.upper() not in convert_type:
            raise CliboaException(
                "Unknown string. One of the followings are allowed [LF, CR, CRLF]"
            )
        return convert_type.get(string.upper())

    @staticmethod
    def extract_columns_with_names(
        input_file,
        output_file,
        remain_column_names,
        enc="utf-8",
    ):
        """
        Extract only the necessary columns from a CSV file and output a new CSV

        Args:
            input_file: Input csv file name
            output_file: Output csv file name
            remain_column_names: Columns which remain
            enc: Encoding
        """
        with open(input_file, mode="r", encoding=enc) as in_f, open(
            output_file, mode="w", encoding=enc
        ) as out_f:
            reader = csv.DictReader(in_f)
            writer = csv.writer(out_f)
            writer.writerow(remain_column_names)
            for row in reader:
                contents = []
                for c in remain_column_names:
                    contents.append(row[c])
                writer.writerow(contents)
            out_f.flush()

    @staticmethod
    def extract_columns_with_numbers(input_file, output_file, remain_column_numbers, enc="utf-8"):
        """
        Extract only the necessary columns from a CSV file and output a new CSV

        Args:
            input_file: Input csv file name
            output_file: Output csv file name
            remain_column_numbers: Column numbers which remain
            enc: Encoding
        """
        with open(input_file, mode="r", encoding=enc) as in_f, open(
            output_file, mode="w", encoding=enc
        ) as out_f:
            reader = csv.reader(in_f)
            writer = csv.writer(out_f)
            for i, row in enumerate(reader):
                contents = []
                for j, r in enumerate(row):
                    if (j + 1) not in remain_column_numbers:
                        continue
                    contents.append(r)
                writer.writerow(contents)
            out_f.flush()

    @staticmethod
    def get_column_names(src, enc="utf-8"):
        """
        Returns csv column names
        """
        columns = []
        with open(src, "r", encoding=enc) as f:
            reader = csv.DictReader(f)
            columns = reader.fieldnames
        return columns



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.base import BaseStep
from cliboa.util.lisboa_log import LisboaLog


class SampleStep(BaseStep):
    """
    For unit test
    """

    def __init__(self):
        super().__init__()
        self._retry_count = 3
        self._memo = None
        self._logger = LisboaLog.get_logger(__name__)

    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def memo(self, memo):
        self._memo = memo

    def execute(self, *args):
        self._logger.info("Start %s" % self.__class__.__name__)
        self._logger.info("Finish %s" % self.__class__.__name__)


class SampleStepSub(SampleStep):
    """
    For unit test
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        self._logger.info("Start %s" % self.__class__.__name__)
        self._logger.info("Finish %s" % self.__class__.__name__)


class SampleCustomStep(BaseStep):
    """
    For unit test
    """

    def __init__(self):
        super().__init__()
        self._password = None
        self._access_key = None
        self._secret_key = None
        self._retry_count = 3

    def password(self, password):
        self._password = password

    def access_key(self, access_key):
        self._access_key = access_key

    def secret_key(self, secret_key):
        self._secret_key = secret_key

    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def execute(self, *args):
        self._logger.info("unit test")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.adapter.azure import BlobServiceAdapter
from cliboa.scenario.azure import BaseAzureBlob
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.constant import StepStatus


class AzureBlobUpload(BaseAzureBlob):
    """
    Upload to Azure Blob storage
    """

    def __init__(self):
        super().__init__()
        self._src_dir = ""
        self._src_pattern = None
        self._dest_dir = None
        self._quit = False

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def quit(self, quit):
        self._quit = quit

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._src_pattern, self._dest_dir]
        )
        valid()

        service = BlobServiceAdapter().get_client(
            account_url=self._account_url,
            account_access_key=self._account_access_key,
            connection_string=self._connection_string,
        )
        files = super().get_target_files(self._src_dir, self._src_pattern)

        if len(files) > 0:
            for f in files:
                path = os.path.join(self._dest_dir, os.path.basename(f))
                blob_client = service.get_blob_client(container=self._container_name, blob=path)
                with open(f, "rb") as data:
                    blob_client.upload_blob(data, overwrite=True)
        else:
            self._logger.info(
                "Files to upload do not exist. File pattern: {}".format(
                    os.path.join(self._src_dir, self._src_pattern)
                )
            )
            if self._quit is True:
                return StepStatus.SUCCESSFUL_TERMINATION



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import errno
from ftplib import FTP, FTP_TLS
from time import sleep

from cliboa.util.lisboa_log import LisboaLog


class FtpAdapter(object):
    """
    Ftp Adaptor
    """

    TIMEOUT_SEC = 30
    RETRY_SEC = 10

    def __init__(
        self,
        host,
        user,
        password,
        timeout=TIMEOUT_SEC,
        retryTimes=3,
        port=21,
        tls=False,
    ):
        """
        Must set whether password or key

        Args:
            host (str): hostname
            user (str): username
            password (str): password
            timeout=30 (int): timeout seconds
            retryTimes=3 (int): retry count
            port=21 (int): port number
            tls=False (bool): use secure connection
        """

        self._host = host
        self._user = user
        self._password = password
        self._timeout = timeout
        self._retryTimes = retryTimes
        self._port = port
        self._tls = tls
        self._logger = LisboaLog.get_logger(__name__)

    def execute(self, obj):
        """
        Execute sftp process.
        Pass a result of cliboa.util.sftp.FtpUtil.{{ function }} for the arguments.

        Args:
            obj (tuple):
                obj[0]: A function to be executed.
                obj[1]: Arguments when function is executed.
        """
        return self._execute(obj[0], obj[1])

    def _execute(self, func, kwargs):

        if not func:
            raise ValueError("Function must not be empty.")

        for _ in range(self._retryTimes):
            try:
                return self._ftp_call(func, **kwargs)
            except Exception as e:
                self._logger.warning(e)
                self._logger.warning(kwargs)

            self._logger.warning("Unexpected error occurred. Retry will start in 10 sec.")
            sleep(FtpAdapter.RETRY_SEC)

        raise IOError(errno.ENOENT, "FTP failed.")

    def _ftp_call(self, func, **kwargs):
        if self._tls:
            with FTP_TLS(host=self._host, timeout=self._timeout) as ftp:
                ftp.set_debuglevel(1)
                ftp.login(user=self._user, passwd=self._password)
                ftp.prot_p()
                return func(ftp=ftp, **kwargs)
        else:
            with FTP() as ftp:
                ftp.set_debuglevel(1)
                ftp.connect(host=self._host, port=self._port, timeout=self._timeout)
                ftp.login(user=self._user, passwd=self._password)
                return func(ftp=ftp, **kwargs)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import configparser
import os
import re
import tempfile
from abc import abstractmethod

from cliboa.conf import env
from cliboa.util.cache import StepArgument
from cliboa.util.exception import FileNotFound, InvalidParameter
from cliboa.util.file import File


class BaseStep(object):
    """
    Base class of all the step classes
    """

    def __init__(self):
        self._step = None
        self._symbol = None
        self._parallel = None
        self._logger = None
        self._listeners = []

    def step(self, step):
        self._step = step

    def symbol(self, symbol):
        self._symbol = symbol

    def parallel(self, parallel):
        self._parallel = parallel

    def logger(self, logger):
        self._logger = logger

    def listeners(self, listeners):
        self._listeners = listeners

    def trigger(self, *args):
        mask = None
        path = os.path.join(env.BASE_DIR, "conf", "cliboa.ini")
        if os.path.exists(path):
            try:
                conf = configparser.ConfigParser()
                conf.read(path, encoding="utf-8")
                mask = conf.get("logging", "mask")
                pattern = re.compile(mask)
            except Exception as e:
                self._logger.warning(e)

        for k, v in self.__dict__.items():
            if mask is not None and pattern.search(k):
                self._logger.info("%s : ****" % k)
            else:
                self._logger.info("%s : %s" % (k, v))
        try:
            for listener in self._listeners:
                listener.before_step(self)

            ret = self.execute(args)

            for listener in self._listeners:
                listener.after_step(self)

            return ret

        except Exception as e:
            for listener in self._listeners:
                listener.error_step(self, e)

            return self._exception_dispatcher(e)
        finally:
            for listener in self._listeners:
                listener.after_completion(self)

    @abstractmethod
    def execute(self, *args):
        pass

    def get_target_files(self, src_dir, src_pattern):
        """
        Search files either with regular expression
        """
        return File().get_target_files(src_dir, src_pattern)

    def get_step_argument(self, name):
        """
        Returns an argument from scenario.yaml definitions
        """
        sa = StepArgument.get(self._symbol)
        if sa:
            return sa.get(name)

    def _property_path_reader(self, src, encoding="utf-8"):
        """
        Returns an resource contents from the path if src starts with "path:",
        returns src if not
        """
        self._logger.warning("DeprecationWarning: Will be removed in the near future")
        if src[:5].upper() == "PATH:":
            fpath = src[5:]
            if os.path.exists(fpath) is False:
                raise FileNotFound(src)
            with open(fpath, mode="r", encoding=encoding) as f:
                return f.read()
        return src

    def _source_path_reader(self, src, encoding="utf-8"):
        """
        Returns an path to temporary file contains content specify in src if src is dict,
        returns src if not
        """
        if src is None:
            return src
        if isinstance(src, dict) and "content" in src:
            with tempfile.NamedTemporaryFile(mode="w", encoding=encoding, delete=False) as fp:
                fp.write(src["content"])
                return fp.name
        elif isinstance(src, dict) and "file" in src:
            if os.path.exists(src["file"]) is False:
                raise FileNotFound(src)
            return src["file"]
        else:
            raise InvalidParameter("The parameter is invalid.")

    def _exception_dispatcher(self, e):
        """
        Handle and dispath CliboaExceptions
        """
        # TODO Currently not doing anything
        raise e



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.adapter.sqlite import SqliteAdapter
from cliboa.util.exception import InvalidParameter, SqliteInvalid
from cliboa.util.lisboa_log import LisboaLog


class EssentialParameters(object):
    """
    Validation for the essential parameters of step class
    """

    def __init__(self, cls_name, param_list):
        """
        Args:
            cls_name: class name which has validation target parameters
            param_list: list of validation target parameters
        """
        self._cls_name = cls_name
        self._param_list = param_list

    def __call__(self):
        for p in self._param_list:
            if not p:
                raise InvalidParameter(
                    "The essential parameter is not specified in %s." % self._cls_name
                )


class SqliteTableExistence(object):
    """
    Validation for the table of sqlite
    """

    def __init__(self, dbname, tblname, returns_bool=False):
        """
        Args:
            dbname: database name
            tblname: table name
            returns_bool: return bool or not
        """
        self._sqlite_adptr = SqliteAdapter()
        self._dbname = dbname
        self._tblname = tblname
        self._returns_bool = returns_bool
        self._logger = LisboaLog.get_logger(__name__)

    def __call__(self):
        try:
            self._sqlite_adptr.connect(self._dbname)
            cur = self._sqlite_adptr.fetch(
                'SELECT name FROM sqlite_master WHERE type="table" AND name="%s"' % self._tblname
            )
            result = cur.fetchall()
            if self._returns_bool is True:
                return True if result else False

            if not result and self._returns_bool is False:
                raise SqliteInvalid("Sqlite table %s not found" % self._tblname)

        finally:
            self._sqlite_adptr.close()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re

from cliboa.adapter.aws import S3Adapter
from cliboa.scenario.aws import BaseS3
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.cache import ObjectStore
from cliboa.util.constant import StepStatus


class S3Download(BaseS3):
    """
    Download from S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None
        self._dest_dir = "."

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(self._access_key, self._secret_key, self._profile)
        client = adapter.get_client()

        p = client.get_paginator("list_objects")
        keys = []
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                path = c.get("Key")
                filename = os.path.basename(path)
                rec = re.compile(self._src_pattern)
                if not rec.fullmatch(filename):
                    continue
                if self._dest_dir:
                    os.makedirs(self._dest_dir, exist_ok=True)
                dest_path = os.path.join(self._dest_dir, filename)
                client.download_file(self._bucket, path, dest_path)
                keys.append(path)

        # cache
        ObjectStore.put(self._step, {"bucket": self._bucket, "keys": keys})


class S3DownloadFileDelete(BaseS3):
    """
    Delete all downloaded files from S3
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        stored = ObjectStore.get(self._symbol)
        bucket = stored.get("bucket")
        keys = stored.get("keys")

        if keys is not None and len(keys) > 0:
            self._region = super().get_step_argument("region")
            self._access_key = super().get_step_argument("access_key")
            self._secret_key = super().get_step_argument("secret_key")
            self._profile = super().get_step_argument("profile")
            self._bucket = super().get_step_argument("bucket")
            self._key = super().get_step_argument("key")
            self._prefix = super().get_step_argument("prefix")
            self._delimiter = super().get_step_argument("delimiter")
            self._src_pattern = super().get_step_argument("src_pattern")

            adapter = S3Adapter(self._access_key, self._secret_key, self._profile)
            client = adapter.get_client()

            for key in keys:
                client.delete_object(Bucket=bucket, Key=key)
                self._logger.info("%s is successfully deleted." % bucket + "/" + key)
        else:
            self._logger.info("No files to delete.")


class S3Delete(BaseS3):
    """
    Delete from S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(self._access_key, self._secret_key, self._profile)
        client = adapter.get_client()

        p = client.get_paginator("list_objects")
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                path = c.get("Key")
                filename = os.path.basename(path)
                rec = re.compile(self._src_pattern)
                if rec.fullmatch(filename) is None:
                    continue
                client.delete_object(Bucket=self._bucket, Key=path)


class S3FileExistsCheck(BaseS3):
    """
    File check in S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(self._access_key, self._secret_key, self._profile)

        p = adapter.get_client().get_paginator("list_objects")
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                filename = os.path.basename(c.get("Key"))
                rec = re.compile(self._src_pattern)
                if not rec.fullmatch(filename):
                    continue
                # The file exist
                self._logger.info("File was found in S3. After process will be processed")
                return None

        # The file does not exist
        self._logger.info("File not found in S3. After process will not be processed")
        return StepStatus.SUCCESSFUL_TERMINATION



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from importlib import import_module

from cliboa.core.manager import JsonScenarioManager, YamlScenarioManager  # noqa
from cliboa.core.strategy import MultiProcExecutor, MultiProcWithConfigExecutor, SingleProcExecutor
from cliboa.util.class_util import ClassUtil
from cliboa.util.parallel_with_config import ParallelWithConfig


class ScenarioManagerFactory(object):
    """
    Create scenario manager instance
    """

    @staticmethod
    def create(cmd_args):
        """
        Create ScenarioManager Instance(YamlScenarioManager or JsonScenarioManager)
        Args:
            cmd_args: Command Line Arguments
        Returns:
            scenario manager instance
        """
        scenario_file_format = cmd_args.format
        scenario_manager_cls = scenario_file_format.capitalize() + "ScenarioManager"
        instance = globals()[scenario_manager_cls]
        return instance(cmd_args)


class StepExecutorFactory(object):
    """
    Create step execution strategy instance
    """

    @staticmethod
    def create(obj):
        """
        Args:
            obj: queue which stores execution target steps
        Returns:
            step execution strategy instance
        """
        if len(obj) > 1:
            return MultiProcExecutor(obj)
        elif len(obj) == 1 and isinstance(obj[0], ParallelWithConfig):
            return MultiProcWithConfigExecutor(obj)
        return SingleProcExecutor(obj)


class CustomInstanceFactory(object):
    """
    Import python module and create instance dynamically

    Return:
        Created instance.
        None: If cls_name was not found in the defined class list.
    """

    @staticmethod
    def create(cls_name):
        ret = ClassUtil().describe_class(cls_name)
        if ret is None:
            return None
        (root, mod_name) = ret
        instance = getattr(import_module(root), mod_name)
        return instance()



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

import os
from importlib import import_module


class Environment:
    """
    setting of cliboa environment values
    """

    def __init__(self):
        """
        Load environment.py in project
        If values are not set in environment.py, default values are used
        """
        env_module = os.environ.get("CLIBOA_ENV")
        if not env_module:
            # Assume executing unit test codes
            mod = import_module("cliboa.common.environment")
        else:
            # Assume executing cfmanager.py
            mod = import_module(env_module)

        for env in dir(mod):
            val = getattr(mod, env)
            setattr(self, env, val)


env = Environment()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.scenario.sqlite import BaseSqlite
from cliboa.scenario.validator import EssentialParameters


class SqliteExport(BaseSqlite):
    def __init__(self):
        super().__init__()
        self._tblname = None
        self._dest_path = None
        self._encoding = "utf-8"
        self._order = []
        self._no_duplicate = False

    def tblname(self, tblname):
        self._tblname = tblname

    def dest_path(self, dest_path):
        self._dest_path = dest_path

    def encoding(self, encoding):
        self._encoding = encoding

    def order(self, order):
        self._order = order

    def no_duplicate(self, no_duplicate):
        self._no_duplicate = no_duplicate

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._dest_path])
        valid()

        dest_dir = os.path.dirname(self._dest_path)
        if dest_dir:
            os.makedirs(dest_dir, exist_ok=True)

        self._sqlite_adptr.connect(self._dbname)
        try:
            self._sqlite_adptr.export_table(
                self._tblname,
                self._dest_path,
                encoding=self._encoding,
                order=self._order,
                no_duplicate=self._no_duplicate,
            )
        finally:
            self._close_database()



import logging
import os

import gnupg

from cliboa.util.exception import CliboaException


class Gpg(object):
    def __init__(self, gnupghome):
        super().__init__()

        if not os.path.exists(gnupghome):
            os.makedirs(gnupghome, exist_ok=True)

        self.log = logging.getLogger(__name__)
        self._gpg = gnupg.GPG(gnupghome=gnupghome)

    def generate_key(self, dest_dir, name_real=None, name_email=None, passphrase=None):
        input_data = self._gpg.gen_key_input(
            name_real=name_real,
            name_email=name_email,
            passphrase=passphrase,
        )

        key = self._gpg.gen_key(input_data)

        public_keys = self._gpg.export_keys(key.fingerprint)
        private_keys = self._gpg.export_keys(
            keyids=key.fingerprint,
            secret=True,
            passphrase=passphrase,
        )

        with open(os.path.join(dest_dir, "public"), "w") as f:
            f.write(public_keys)

        with open(os.path.join(dest_dir, "private"), "w") as f:
            f.write(private_keys)

    def import_key(self, key_path):
        with open(key_path) as f:
            key_data = f.read()
            import_result = self._gpg.import_keys(key_data)

        self.log.info("Import key count: %s " % import_result.count)

    def trust_key(self, key_path, trust_level):
        keys = self._gpg.scan_keys(key_path)
        for key in keys:
            self._gpg.trust_keys(key["fingerprint"], trust_level)

    def encrypt(self, src_path, dest_path, recipients, passphrase=None, always_trust=False):
        with open(src_path, "rb") as f:
            status = self._gpg.encrypt_file(
                file=f,
                recipients=recipients,
                always_trust=always_trust,
                passphrase=passphrase,
                output=dest_path + ".gpg",
            )

        if status.ok is False:
            self.log.error(status.stderr)
            raise CliboaException("Failed to encrypt gpg.")

        self.log.info(status.ok)
        self.log.info(status.status)

    def decrypt(self, src_path, dest_path, passphrase=None, always_trust=False):
        with open(src_path, "rb") as f:
            status = self._gpg.decrypt_file(
                file=f,
                always_trust=always_trust,
                passphrase=passphrase,
                output=dest_path,
            )

        if status.ok is False:
            self.log.error(status.stderr)
            raise CliboaException("Failed to decrypt gpg.")

        self.log.info(status.ok)
        self.log.info(status.status)



# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import boto3
from boto3.session import Session

from cliboa.util.exception import InvalidParameter


class S3Adapter(object):
    """
    Adapter of AWS S3
    """

    def __init__(self, access_key: str = None, secret_key: str = None, profile: str = None):

        if (access_key and secret_key) and profile:
            raise InvalidParameter(
                "Either access_key and secret_key or profile path can be specified."
            )

        self._access_key = access_key
        self._secret_key = secret_key
        self._profile = profile

    def get_client(self):
        """
        Get s3 client
        """
        if self._profile:
            return Session(profile_name=self._profile).client("s3")
        elif self._access_key and self._secret_key:
            return boto3.client(
                "s3",
                aws_access_key_id=self._access_key,
                aws_secret_access_key=self._secret_key,
            )
        else:
            return boto3.client("s3")

    def get_resource(self):
        """
        Get s3 resource
        """
        if self._profile:
            return Session(profile_name=self._profile).resource("s3")
        elif self._access_key and self._secret_key:
            session = Session(
                aws_access_key_id=self._access_key,
                aws_secret_access_key=self._secret_key,
            )
            return session.resource("s3")
        else:
            return boto3.resource("s3")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.adapter.sqlite import SqliteAdapter
from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.constant import StepStatus


class BaseSqlite(BaseStep):
    """
    Base class of all the sqlite classes
    """

    def __init__(self):
        super().__init__()
        self._sqlite_adptr = SqliteAdapter()
        self._dbname = None
        self._vacuum = False

    def dbname(self, dbname):
        self._dbname = dbname

    def vacuum(self, vacuum):
        self._vacuum = vacuum

    def execute(self, *args):
        # essential parameters check
        param_valid = EssentialParameters(self.__class__.__name__, [self._dbname])
        param_valid()

    def _dict_factory(self, cursor, row):
        d = {}
        for i, col in enumerate(cursor.description):
            d[col[0]] = row[i]
        return d

    def _close_database(self):
        """
        Disconnect sqlite database (execute vacuum if necessary)
        """
        self._sqlite_adptr.close()
        if self._vacuum is True:
            try:
                self._sqlite_adptr.connect(self._dbname)
                self._sqlite_adptr.execute("VACUUM")
            finally:
                self._sqlite_adptr.close()


class SqliteTransaction(BaseSqlite):
    """
    Base class of sqlite transaction
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        """
        Execute db connection, commit, close, vacuum

        Args:
            args[0]: db operation function to execute as transaction
        """

        # When args is not specified, just call validation of base class
        if not args:
            super().execute()
            return StepStatus.ABNORMAL_TERMINATION
        func = args[0]

        # TODO
        # Add to handle to select DB transaction seperation level
        self._logger.info("Start DB Transaction")
        self._sqlite_adptr.connect(self._dbname)
        try:
            func()
            self._sqlite_adptr.commit()
        finally:
            super()._close_database()
        self._logger.info("Finish DB Transaction")


class SqliteQueryExecute(SqliteTransaction):
    """
    Execute only row query of insert, update or delete
    """

    def __init__(self):
        super().__init__()
        self._raw_query = None

    def raw_query(self, raw_query):
        self._raw_query = raw_query

    def execute(self, *args):
        param_valid = EssentialParameters(self.__class__.__name__, [self._raw_query])
        param_valid()

        def func():
            self._sqlite_adptr.execute(self._raw_query)

        super().execute(func)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

import csv
import hashlib
import os
import re
from datetime import datetime

import jsonlines
import pandas

from cliboa.adapter.sqlite import SqliteAdapter
from cliboa.core.validator import EssentialParameters
from cliboa.scenario.transform.file import FileBaseTransform
from cliboa.util.csv import Csv
from cliboa.util.exception import FileNotFound, InvalidCount, InvalidParameter
from cliboa.util.file import File
from cliboa.util.string import StringUtil


class CsvColumnHash(FileBaseTransform):
    """
    Hash(SHA256) specific columns from csv file.
    """

    def __init__(self):
        super().__init__()
        self._columns = []

    def columns(self, columns):
        self._columns = columns

    def _stringToHash(self, string):
        return hashlib.sha256(string.encode()).hexdigest()

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._columns],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        for df in tfr:
            for c in self._columns:
                df[c] = df[c].apply(self._stringToHash)
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvColumnExtract(FileBaseTransform):
    """
    Remove specific columns from csv file.
    """

    def __init__(self):
        super().__init__()
        self._columns = None
        self._column_numbers = None

    def columns(self, columns):
        self._columns = columns

    def column_numbers(self, column_numbers):
        self._column_numbers = column_numbers

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        if not self._columns and not self._column_numbers:
            raise InvalidParameter("Specifying either 'column' or 'column_numbers' is essential.")
        if self._columns and self._column_numbers:
            raise InvalidParameter("Cannot specify both 'column' and 'column_numbers'.")

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        if self._columns:
            Csv.extract_columns_with_names(fi, fo, self._columns)
        elif self._column_numbers:
            if isinstance(self._column_numbers, int) is True:
                remain_column_numbers = [self._column_numbers]
            else:
                column_numbers = self._column_numbers.split(",")
                remain_column_numbers = [int(n) for n in column_numbers]
            Csv.extract_columns_with_numbers(fi, fo, remain_column_numbers)


class CsvColumnDelete(FileBaseTransform):
    """
    Delete specific columns from csv file.
    """

    def __init__(self):
        super().__init__()
        self._regex_pattern = None

    def regex_pattern(self, regex_pattern):
        self._regex_pattern = regex_pattern

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        if not self._regex_pattern:
            raise InvalidParameter("'regex_pattern' is essential.")

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        pattern = re.compile(self._regex_pattern)
        for df in tfr:
            for column in df.columns.values:
                if pattern.fullmatch(column):
                    df = df.drop(column, axis=1)
                # output an empty file when all columns are deleted
                if df.empty:
                    df = df.dropna(how="all")
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvValueExtract(FileBaseTransform):
    """
    Extract a specific column from a CSV file and then replace it with a regular expression.
    """

    def __init__(self):
        super().__init__()
        self._column_regex_pattern = None

    def column_regex_pattern(self, column_regex_pattern):
        self._column_regex_pattern = column_regex_pattern

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._column_regex_pattern],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)
        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        with open(fi, mode="rt") as i:
            reader = csv.DictReader(i)
            with open(fo, mode="wt") as o:
                writer = csv.DictWriter(o, reader.fieldnames)
                writer.writeheader()
                for line in reader:
                    for column, regex_pattern in self._column_regex_pattern.items():
                        if re.search(regex_pattern, line[column]):
                            line[column] = re.search(regex_pattern, line[column]).group()
                        else:
                            line[column] = ""
                    writer.writerow(dict(line))


class CsvColumnConcat(FileBaseTransform):
    """
    Concat specific columns from csv file.
    """

    def __init__(self):
        super().__init__()
        self._columns = None
        self._dest_column_name = None
        self._sep = ""

    def columns(self, columns):
        self._columns = columns

    def dest_column_name(self, dest_column_name):
        self._dest_column_name = dest_column_name

    def sep(self, sep):
        self._sep = sep

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._dest_column_name],
        )
        valid()

        if len(self._columns) < 2 or type(self._columns) is not list:
            raise InvalidParameter("'columns' must 2 or more lengths")

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        dest_str = None
        for df in tfr:
            for c in self._columns:
                if dest_str is None:
                    dest_str = df[c].astype(str)
                else:
                    dest_str = dest_str + self._sep + df[c].astype(str)
                df = df.drop(columns=[c])
            df[self._dest_column_name] = dest_str
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvMergeExclusive(FileBaseTransform):
    """
    Compare specific columns each file.
    If matched, exclude rows.
    """

    def __init__(self):
        super().__init__()
        self._src_column = None
        self._target_compare_path = None
        self._target_column = None

    def src_column(self, src_column):
        self._src_column = src_column

    def target_compare_path(self, target_compare_path):
        self._target_compare_path = target_compare_path

    def target_column(self, target_column):
        self._target_column = target_column

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [
                self._src_dir,
                self._src_pattern,
                self._src_column,
                self._target_compare_path,
                self._target_column,
            ],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        target = super().get_target_files(
            os.path.dirname(self._target_compare_path),
            os.path.basename(self._target_compare_path),
        )
        self.check_file_existence(target)

        header = pandas.read_csv(self._target_compare_path, nrows=0)
        if self._target_column not in header:
            raise KeyError(
                "Target Compare file does not exist target column [%s]." % self._target_column
            )

        df_target = pandas.read_csv(self._target_compare_path, usecols=[self._target_column])
        self.df_target_list = df_target[self._target_column].values.tolist()

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        header = pandas.read_csv(fi, dtype=str, encoding=self._encoding, nrows=0)
        try:
            header[self._src_column].values.tolist()
        except KeyError:
            raise KeyError("Src file does not exist target column [%s]." % self._target_column)

        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(fi, chunksize=chunksize)
        for df in tfr:
            df = df[~df[self._src_column].isin(self.df_target_list)]
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class ColumnLengthAdjust(FileBaseTransform):
    """
    Adjust csv (tsv) column to maximum length
    """

    def __init__(self):
        super().__init__()
        self._adjust = {}

    def adjust(self, adjust):
        self._adjust = adjust

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._adjust],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_writers(files, encoding=self._encoding)

        for fi, fo in super().io_writers(files, encoding=self._encoding):
            reader = csv.DictReader(fi)
            writer = csv.DictWriter(fo, reader.fieldnames)
            writer.writeheader()

            for row in reader:
                for k, v in self._adjust.items():
                    f1 = row.get(k)
                    if len(f1) > v:
                        row[k] = f1[:v]
                writer.writerow(row)
            fo.flush()


class CsvMerge(FileBaseTransform):
    """
    Merge two csv files
    """

    def __init__(self):
        super().__init__()
        self._src1_pattern = None
        self._src2_pattern = None

    def src1_pattern(self, src1_pattern):
        self._src1_pattern = src1_pattern

    def src2_pattern(self, src2_pattern):
        self._src2_pattern = src2_pattern

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [
                self._src_dir,
                self._src1_pattern,
                self._src2_pattern,
                self._dest_dir,
                self._dest_name,
            ],
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        target1_files = File().get_target_files(self._src_dir, self._src1_pattern)
        target2_files = File().get_target_files(self._src_dir, self._src2_pattern)
        if len(target1_files) == 0:
            raise InvalidCount(
                "An input file %s does not exist." % os.path.join(self._src_dir, self._src1_pattern)
            )
        elif len(target2_files) == 0:
            raise InvalidCount(
                "An input file %s does not exist." % os.path.join(self._src_dir, self._src2_pattern)
            )
        elif len(target1_files) > 1:
            self._logger.error("Hit target files %s" % target1_files)
            raise InvalidCount("Input files must be only one.")
        elif len(target2_files) > 1:
            self._logger.error("Hit target files %s" % target2_files)
            raise InvalidCount("Input files must be only one.")

        self._logger.info("Merge %s and %s." % (target1_files[0], target2_files[0]))

        chunk_size_handling(self._read_csv_func, target1_files, target2_files)

    def _read_csv_func(self, chunksize, target1_files, target2_files):
        # Used in chunk_size_handling
        first_write = True
        tfr1 = pandas.read_csv(
            os.path.join(self._src_dir, target1_files[0]),
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        for df in tfr1:
            df.to_csv(
                os.path.join(self._dest_dir, self._dest_name),
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False
        tfr2 = pandas.read_csv(
            os.path.join(self._src_dir, target2_files[0]),
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        for df in tfr2:
            df.to_csv(
                os.path.join(self._dest_dir, self._dest_name),
                encoding=self._encoding,
                header=False,
                index=False,
                mode="a",
            )


class CsvColumnSelect(FileBaseTransform):
    """
    Select columns in Csv file in specified order
    """

    def __init__(self):
        super().__init__()
        self._column_order = None

    def column_order(self, column_order):
        self._column_order = column_order

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._column_order],
        )
        valid()

        files = File().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            raise FileNotFound("No files are found.")
        self._logger.info("Files found %s" % files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )
        for df in tfr:
            if set(self._column_order) - set(df.columns.values):
                raise InvalidParameter(
                    "column_order define not included target file's column : %s"
                    % (set(self._column_order) - set(df.columns.values))
                )
            df = df.loc[:, self._column_order]
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvConcat(FileBaseTransform):
    """
    Concat csv files
    """

    def __init__(self):
        super().__init__()
        self._src_filenames = None

    def src_filenames(self, src_filenames):
        self._src_filenames = src_filenames

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._dest_dir, self._dest_name]
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        if not self._src_pattern and not self._src_filenames:
            raise InvalidParameter(
                "Specifying either 'src_pattern' or 'src_filenames' is essential."
            )
        if self._src_pattern and self._src_filenames:
            raise InvalidParameter("Cannot specify both 'src_pattern' and 'src_filenames'.")

        if self._src_pattern:
            files = File().get_target_files(self._src_dir, self._src_pattern)
        else:
            files = []
            for file in self._src_filenames:
                files.append(os.path.join(self._src_dir, file))

        if len(files) == 0:
            raise FileNotFound("No files are found.")
        elif len(files) == 1:
            self._logger.warning("Two or more input files are required.")

        # Create output headers to conform to the concat specification.
        file_1 = files[0]
        output_header = pandas.read_csv(file_1, dtype=str, encoding=self._encoding, nrows=0)
        for file in files[1:]:
            output_header = pandas.concat(
                [
                    output_header,
                    pandas.read_csv(file, dtype=str, encoding=self._encoding, nrows=0),
                ]
            )

        chunk_size_handling(self._read_csv_func, files, output_header)

    def _read_csv_func(self, chunksize, files, output_header):
        # Used in chunk_size_handling
        first_write = True
        for file in files:
            tfr = pandas.read_csv(
                file,
                dtype=str,
                encoding=self._encoding,
                chunksize=chunksize,
            )
            for df in tfr:
                # Change the header order to the one you plan to output.
                df = pandas.concat([output_header, df])
                df.to_csv(
                    os.path.join(self._dest_dir, self._dest_name),
                    encoding=self._encoding,
                    header=True if first_write else False,
                    index=False,
                    mode="w" if first_write else "a",
                )
                first_write = False


class CsvConvert(FileBaseTransform):
    """
    Change csv format
    """

    def __init__(self):
        super().__init__()
        self._headers = []
        self._headers_existence = True
        self._add_headers = None
        self._before_format = "csv"
        self._before_enc = self._encoding
        self._after_format = None
        self._after_enc = None
        self._after_nl = "LF"
        self._reader_quote = "QUOTE_MINIMAL"
        self._quote = "QUOTE_MINIMAL"

    def headers(self, headers):
        self._headers = headers

    def headers_existence(self, headers_existence):
        self._headers_existence = headers_existence

    def add_headers(self, add_headers):
        self._add_headers = add_headers

    def before_format(self, before_format):
        self._before_format = before_format

    def before_enc(self, before_enc):
        self._before_enc = before_enc

    def after_format(self, after_format):
        self._after_format = after_format

    def after_enc(self, after_enc):
        self._after_enc = after_enc

    def after_nl(self, after_nl):
        self._after_nl = after_nl

    def reader_quote(self, reader_quote):
        self._reader_quote = reader_quote

    def quote(self, quote):
        self._quote = quote

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._before_format, self._before_enc],
        )
        valid()

        if self._after_format is None:
            self._after_format = self._before_format

        if self._after_enc is None:
            self._after_enc = self._before_enc

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, ext=self._after_format, func=self.convert)

    def convert(self, fi, fo):
        with open(fi, mode="rt", encoding=self._before_enc) as i:
            reader = csv.reader(
                i,
                delimiter=Csv.delimiter_convert(self._before_format),
                quoting=Csv.quote_convert(self._reader_quote),
            )
            with open(fo, mode="wt", newline="", encoding=self._after_enc) as o:
                writer = csv.writer(
                    o,
                    delimiter=Csv.delimiter_convert(self._after_format),
                    quoting=Csv.quote_convert(self._quote),
                    lineterminator=Csv.newline_convert(self._after_nl),
                )

                for i, line in enumerate(reader):
                    if i == 0:
                        if self._headers_existence is False:
                            continue
                        if self._add_headers:
                            writer.writerow(self._add_headers)
                            writer.writerow(line)
                        else:
                            writer.writerow(self._replace_headers(line))
                    else:
                        writer.writerow(line)

    def _replace_headers(self, old_headers):
        """
        Replace old headers to new headers
        """
        if not self._headers:
            return old_headers

        converter = {}
        for headers in self._headers:
            for k, v in headers.items():
                converter[k] = v

        new_headers = []
        for oh in old_headers:
            r = converter.get(oh)
            new_headers.append(r if r is not None else oh)

        return new_headers


class CsvSort(FileBaseTransform):
    """
    Sort csv.
    """

    def __init__(self):
        super().__init__()
        self._order = []
        self._quote = "QUOTE_MINIMAL"
        self._no_duplicate = False

    def order(self, order):
        self._order = order

    def quote(self, quote):
        self._quote = quote

    def no_duplicate(self, no_duplicate):
        self._no_duplicate = no_duplicate

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._order, self._src_dir, self._src_pattern, self._dest_dir],
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        dbname = ".%s_%s.db" % (ymd_hms, StringUtil().random_str(8))
        tblname = "temp_table"

        sqlite = SqliteAdapter()
        sqlite.connect(dbname)
        try:
            for file in files:
                _, filename = os.path.split(file)
                dest_file = os.path.join(self._dest_dir, filename)

                sqlite.import_table(file, tblname, encoding=self._encoding)
                sqlite.export_table(
                    tblname,
                    dest_file,
                    quoting=Csv.quote_convert(self._quote),
                    encoding=self._encoding,
                    order=self._order,
                    no_duplicate=self._no_duplicate,
                )
            sqlite.close()
        finally:
            os.remove(dbname)


class CsvToJsonl(FileBaseTransform):
    """
    Transform csv to jsonlines.
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, ext="jsonl", func=self.convert)

    def convert(self, fi, fo):
        with open(fi, mode="r", encoding=self._encoding, newline="") as i, jsonlines.open(
            fo, mode="w"
        ) as writer:
            reader = csv.DictReader(i)
            for row in reader:
                writer.write(row)


class CsvColumnCopy(FileBaseTransform):
    """
    Copy column data (new or overwrite)
    """

    def __init__(self):
        super().__init__()
        self._src_column = None
        self._dest_column = None

    def src_column(self, src_column):
        self._src_column = src_column

    def dest_column(self, dest_column):
        self._dest_column = dest_column

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._dest_dir, self._src_column, self._dest_column],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        header = pandas.read_csv(fi, dtype=str, encoding=self._encoding, nrows=0)
        if self._src_column not in header:
            raise KeyError("Copy source column does not exist in file. [%s]" % self._src_column)

        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )

        for df in tfr:
            df[self._dest_column] = df[self._src_column]
            df.to_csv(
                fo,
                encoding=self._encoding,
                header=True if first_write else False,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


class CsvColumnReplace(FileBaseTransform):
    """
    Replace matching regular expression values for a specific column from a csv file.
    """

    def __init__(self):
        super().__init__()
        self._column = None
        self._regex_pattern = None
        self._rep_str = None
        self._regex_compile = None

    def column(self, column):
        self._column = column

    def regex_pattern(self, regex_pattern):
        self._regex_pattern = regex_pattern

    def rep_str(self, rep_str):
        self._rep_str = rep_str

    def _replace_string(self, string):
        return self._regex_compile.sub(self._rep_str, string)

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._column],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        if self._rep_str is None:
            raise InvalidParameter("The converted string is not defined in yaml file: rep_str")
        if self._regex_pattern is None:
            raise InvalidParameter(
                "The conversion pattern is not defined in yaml file: regex_pattern"
            )
        self._regex_compile = re.compile(self._regex_pattern)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        header = pandas.read_csv(fi, dtype=str, encoding=self._encoding, nrows=0)
        if self._column not in header:
            raise KeyError("Replace source column does not exist in file. [%s]" % self._column)

        chunk_size_handling(self._read_csv_func, fi, fo)

    def _read_csv_func(self, chunksize, fi, fo):
        # Used in chunk_size_handling
        first_write = True
        tfr = pandas.read_csv(
            fi,
            dtype=str,
            encoding=self._encoding,
            chunksize=chunksize,
        )

        for df in tfr:
            df[self._column] = df[self._column].apply(self._replace_string)
            df.to_csv(
                fo,
                header=True if first_write else False,
                encoding=self._encoding,
                index=False,
                mode="w" if first_write else "a",
            )
            first_write = False


def chunk_size_handling(read_csv_func, *args, **kwd):
    """
    Processing to avoid memory errors in pandas's read_csv.
    Use this function when you want to do the same handling when extending cliboa.
    """
    chunksize = 1024 * 1024
    while 0 < chunksize:
        try:
            read_csv_func(chunksize, *args, **kwd)
            break
        except MemoryError as error:
            if chunksize <= 1:
                raise error
            chunksize //= 2



# flake8: noqa
#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from .extract.aws import S3Delete, S3Download, S3DownloadFileDelete, S3FileExistsCheck
from .extract.azure import AzureBlobDownload
from .extract.ftp import FtpDownload, FtpDownloadFileDelete
from .extract.gcp import (
    BigQueryRead,
    FirestoreDocumentDownload,
    GcsDownload,
    GcsDownloadFileDelete,
    GcsFileExistsCheck,
)
from .extract.http import HttpDownload, HttpDownloadViaBasicAuth
from .extract.mysql import MysqlRead
from .extract.postgres import PostgresqlRead
from .extract.sftp import SftpDelete, SftpDownload, SftpDownloadFileDelete, SftpFileExistsCheck
from .extract.sqlite import SqliteExport
from .load.aws import S3Upload
from .load.azure import AzureBlobUpload
from .load.gcp import BigQueryCopy, BigQueryWrite, FirestoreDocumentCreate, GcsUpload
from .load.mysql import MysqlWrite
from .load.postgres import PostgresqlWrite
from .load.sftp import SftpUpload
from .load.sqlite import SqliteImport
from .sqlite import SqliteQueryExecute
from .transform.csv import (
    ColumnLengthAdjust,
    CsvColumnConcat,
    CsvColumnCopy,
    CsvColumnDelete,
    CsvColumnExtract,
    CsvColumnHash,
    CsvColumnReplace,
    CsvConcat,
    CsvConvert,
    CsvMerge,
    CsvSort,
    CsvToJsonl,
    CsvValueExtract,
)
from .transform.file import (
    DateFormatConvert,
    ExcelConvert,
    FileArchive,
    FileCompress,
    FileConvert,
    FileDecompress,
    FileDivide,
    FileRename,
)
from .transform.gpg import GpgDecrypt, GpgEncrypt, GpgGenerateKey
from .transform.json import JsonlAddKeyValue, JsonlToCsv, JsonlToCsvBase
from .transform.system import ExecuteShellScript



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import bz2
import csv
import gzip
import os
import re
import shutil
import tarfile
import tempfile
import zipfile

import pandas

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.base import BaseStep
from cliboa.scenario.extras import ExceptionHandler
from cliboa.util.date import DateUtil
from cliboa.util.exception import CliboaException, FileNotFound, InvalidParameter
from cliboa.util.file import File


class FileBaseTransform(BaseStep, ExceptionHandler):
    """
    Base class of file transform classes

    Basically transform class of Cliboa is that find files,
    do something, and output transformed files.
    When output files are created, name of the files will be the same name with the input files
    (original input files will be changed to the transformed files).
    If you would not like to remove the original files,
    give a path to the "dest_dir" for output directory.
    (If a non-existent directory path is specified, the directory is automatically created.)

    Note:
    Output files are not always be the same name with the input file names.
    See documentation for individual classes for details.
    """

    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._dest_dir = None
        self._dest_name = None
        self._encoding = "utf-8"
        self._nonfile_error = False
        self._force_continue = False

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def dest_name(self, dest_name):
        self._dest_name = dest_name

    def encoding(self, encoding):
        self._encoding = encoding

    def nonfile_error(self, nonfile_error):
        self._nonfile_error = nonfile_error

    def force_continue(self, force_continue):
        self._force_continue = force_continue

    def execute(self, *args):
        pass

    def check_file_existence(self, files):
        """
        Check whether files exist.
        If no files, the scenario will continue or an error is raised,
        depends on parameter[nonfile_error].
        """
        if len(files) == 0:
            if self._nonfile_error is True:
                raise FileNotFound("No files are found.")
            else:
                self._logger.info("No files are found. Nothing to do.")
                return
        self._logger.info("Files found %s" % files)

    def check_output_path(self, input_path, ext):
        root, name = os.path.split(input_path)

        if ext:
            if ext.startswith("."):
                output_name = os.path.splitext(name)[0] + ext
            else:
                output_name = os.path.splitext(name)[0] + "." + ext
        else:
            output_name = name

        if self._dest_dir:
            os.makedirs(self._dest_dir, exist_ok=True)
            output_dir = self._dest_dir
        else:
            output_dir = root

        output_path = os.path.join(output_dir, output_name)

        fd, temp_file = tempfile.mkstemp()
        os.close(fd)
        return output_path, temp_file

    def overwrite_output_path(self, input_path, output_path, temp_file):
        if input_path == output_path:
            os.remove(input_path)
        shutil.move(temp_file, output_path)

    def io_files(self, iterable, ext=None, func=None):
        """
        Iteration which returns input and output path.
        If the parameter "dest_dir" was given, the output file will be created under
        the given directory and returns input and output path.
        If not, the output file will be created to the same directory to the input file.

        Arguments:
            iterable (list): Input file list
            ext=None (str): Set an extension for output file,
                            if input and output extension would like to be changed.
                            "." is not necessary.
        """
        for input_path in iterable:
            output_path, temp_file = self.check_output_path(input_path, ext)

            try:
                func(input_path, temp_file)
            except Exception as e:
                if self._force_continue is True:
                    self.handle_error(e, input_path)
                else:
                    raise e

            self.overwrite_output_path(input_path, output_path, temp_file)

    def io_writers(self, iterable, mode="t", encoding="utf-8", ext=None):
        """
        Iteration which returns input and output writer.
        If the parameter "dest_dir" was given, the output file will be created under
        the given directory and returns input and output writer.
        If not, the output file will be created to the same directory to the input file.

        Arguments:
            iterable (list): Input file list
            mode="t" (str): Mode to open input files, opening file with text mode by default.
            encoding="utf-8" (str): File encoding. It will be ignored when mode is "b"
            ext=None (str): Set an extension for output file,
                            if input and output extension would like to be changed.
                            "." is not necessary.
        """
        if not "t" == mode and not "b" == mode:
            raise InvalidParameter("Unknown mode. One of the following is allowed [t, b]")

        if "b" in mode:
            encoding = None

        for input_path in iterable:
            output_path, temp_file = self.check_output_path(input_path, ext)

            if mode == "t":
                with open(input_path, mode="r", encoding=encoding, newline="") as i, open(
                    temp_file, mode="w", encoding=encoding, newline=""
                ) as o:
                    yield i, o
            elif mode == "b":
                with open(input_path, mode="rb") as i, open(temp_file, mode="wb") as o:
                    yield i, o

            self.overwrite_output_path(input_path, output_path, temp_file)


class FileDecompress(FileBaseTransform):
    """
    Decompress the specified file
    """

    def __init__(self):
        super().__init__()
        self._chunk_size = None
        self._password = None

    def chunk_size(self, chunk_size):
        self._chunk_size = chunk_size

    def password(self, password):
        self._password = password

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        for f in files:
            _, ext = os.path.splitext(f)
            if ext == ".zip":
                self._logger.info("Decompress zip file %s" % f)
                if self._password is not None:
                    pwd = self._password.encode(self._encoding)
                else:
                    pwd = self._password
                with zipfile.ZipFile(f) as zp:
                    zp.extractall(
                        self._dest_dir if self._dest_dir is not None else self._src_dir, pwd=pwd
                    ),
            elif ext == ".tar":
                self._logger.info("Decompress tar file %s" % f)
                with tarfile.open(f, "r:*") as tf:
                    tf.extractall(self._dest_dir if self._dest_dir is not None else self._src_dir)
            elif ext == ".bz2":
                self._logger.info("Decompress bz2 file %s" % f)
                dcom_name = os.path.splitext(os.path.basename(f))[0]
                decom_path = (
                    os.path.join(self._dest_dir, dcom_name)
                    if self._dest_dir is not None
                    else os.path.join(self._src_dir, dcom_name)
                )
                with bz2.open(f, mode="rb") as i, open(decom_path, mode="wb") as o:
                    while True:
                        buf = i.read(self._chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            elif ext == ".gz":
                self._logger.info("Decompress gz file %s" % f)
                dcom_name = os.path.splitext(os.path.basename(f))[0]
                if self._dest_dir:
                    os.makedirs(self._dest_dir, exist_ok=True)
                    decom_path = os.path.join(self._dest_dir, dcom_name)
                else:
                    decom_path = os.path.join(self._src_dir, dcom_name)
                with gzip.open(f, "rb") as i, open(decom_path, "wb") as o:
                    while True:
                        buf = i.read(self._chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            else:
                raise CliboaException("Unmatched any available decompress type %s" % f)


class FileCompress(FileBaseTransform):
    """
    Compress files
    """

    def __init__(self):
        super().__init__()
        self._format = None
        self._chunk_size = 1048576

    def format(self, format):
        self._format = format.lower()

    def chunk_size(self, chunk_size):
        self._chunk_size = chunk_size

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._format],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        if self._dest_dir:
            os.makedirs(self._dest_dir, exist_ok=True)
            dir = self._dest_dir
        else:
            dir = self._src_dir
        for f in files:
            if self._format == "zip":
                self._logger.info("Compress file %s to zip." % f)
                with zipfile.ZipFile(
                    os.path.join(dir, (os.path.basename(f) + ".zip")),
                    "w",
                    zipfile.ZIP_DEFLATED,
                ) as o:
                    o.write(f, arcname=os.path.basename(f))
            elif self._format in ("gz", "gzip"):
                self._logger.info("Compress file %s to gzip." % f)
                com_path = os.path.join(dir, (os.path.basename(f) + ".gz"))
                with open(f, "rb") as i, gzip.open(com_path, "wb") as o:
                    while True:
                        buf = i.read(self._chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)
            elif self._format in ("bz2", "bzip2"):
                self._logger.info("Compress file %s to bzip2." % f)
                com_path = os.path.join(dir, (os.path.basename(f) + ".bz2"))
                with open(f, "rb") as i, bz2.open(com_path, "wb") as o:
                    while True:
                        buf = i.read(self._chunk_size)
                        if buf == b"":
                            break
                        o.write(buf)


class DateFormatConvert(FileBaseTransform):
    """
    Convert csv (tsv) date field columns to another date field format columns
    """

    def __init__(self):
        super().__init__()
        self._columns = []
        self._formatter = None

    def columns(self, columns):
        self._columns = columns

    def formatter(self, formatter):
        self._formatter = formatter

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._columns, self._formatter],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        _, ext = os.path.splitext(files[0])
        if ext == ".csv":
            delimiter = ","
        elif ext == ".tsv":
            delimiter = "\t"

        for ins, ous in super().io_writers(files, encoding=self._encoding):
            reader = csv.DictReader(ins, delimiter=delimiter)
            writer = csv.DictWriter(ous, reader.fieldnames)
            writer.writeheader()
            date_util = DateUtil()
            for row in reader:
                for column in self._columns:
                    r = row.get(column)
                    if not r:
                        continue
                    row[column] = date_util.convert_date_format(r, self._formatter)
                writer.writerow(row)


class ExcelConvert(FileBaseTransform):
    """
    Convert excel to other format
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        # TODO Currently only excel to csv is supported.
        super().io_files(files, ext="csv", func=self.convert)

    def convert(self, fi, fo):
        self._logger.info("Convert %s to %s" % (fi, fo))
        df = pandas.read_excel(fi)
        df.to_csv(fo, encoding=self._encoding)


class FileDivide(FileBaseTransform):
    """
    Divide a file to plural files
    """

    def __init__(self):
        super().__init__()
        self._divide_rows = None
        self._header = False

    def divide_rows(self, divide_rows):
        self._divide_rows = divide_rows

    def header(self, header):
        self._header = header

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._divide_rows],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        for file in files:
            fname = os.path.basename(file)

            px = ""
            if fname.startswith("."):
                fname = fname[1:]
                px = "."

            if "." in fname:
                nameonly, ext = fname.split(".", 1)
                ext = "." + ext
            else:
                nameonly = fname
                ext = ""

            if self._header:
                with open(file, encoding=self._encoding) as i:
                    self._header_row = i.readline()

            row = self._ifile_reader(file)
            newfilename = px + nameonly + ".%s" + ext

            if self._dest_dir:
                os.makedirs(self._dest_dir, exist_ok=True)
                dest_dir = self._dest_dir
            else:
                dest_dir = self._src_dir
            has_left = True
            index = 1
            while has_left:
                ofile_path = os.path.join(dest_dir, newfilename % str(index))
                has_left = self._ofile_generator(ofile_path, row)
                index = index + 1

    def _ifile_reader(self, filepath):
        with open(filepath, encoding=self._encoding) as i:
            if self._header is True:
                i.readline()
            for line in i:
                yield line

    def _ofile_generator(self, filepath, row):
        left = False
        written = False
        with open(filepath, mode="w", encoding=self._encoding) as o:
            if self._header is True:
                o.write(self._header_row)
            for i, line in enumerate(row):
                written = True
                o.write(line)
                if i + 1 >= self._divide_rows:
                    left = True
                    break
        if written is False:
            os.remove(filepath)
        return left


class FileRename(FileBaseTransform):
    """
    Change file names with adding either prefix or suffix.
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._suffix = ""
        self._regex_pattern = None
        self._rep_str = None
        self._ext = ""

    def prefix(self, prefix):
        self._prefix = prefix

    def suffix(self, suffix):
        self._suffix = suffix

    def regex_pattern(self, regex_pattern):
        self._regex_pattern = regex_pattern

    def rep_str(self, rep_str):
        self._rep_str = rep_str

    def ext(self, ext):
        self._ext = ext

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        for file in files:
            dirname = os.path.dirname(file)
            basename = os.path.basename(file)

            px = ""
            if basename.startswith("."):
                basename = basename[1:]
                px = "."

            if "." in basename:
                nameonly, extension = basename.split(".", 1)
                extension = "." + extension
            else:
                nameonly = basename
                extension = ""

            if self._regex_pattern is not None and self._rep_str is not None:
                nameonly = re.sub(self._regex_pattern, self._rep_str, nameonly)
            elif self._regex_pattern is not None:
                raise InvalidParameter("The converted string is not defined in yaml file: dest_str")
            elif self._rep_str is not None:
                raise InvalidParameter(
                    "The conversion pattern is not defined in yaml file: regex_pattern"
                )

            if self._ext:
                extension = "." + self._ext

            newfilename = self._prefix + px + nameonly + self._suffix + extension
            newfilepath = os.path.join(dirname, newfilename)
            os.rename(file, newfilepath)
            self._logger.info("File name changed %s -> %s" % (file, newfilepath))


class FileConvert(FileBaseTransform):
    """
    Convert file encoding
    """

    def __init__(self):
        super().__init__()
        self._encoding_from = None
        self._encoding_to = None
        self._errors = None

    def encoding_from(self, encoding_from):
        self._encoding_from = encoding_from

    def encoding_to(self, encoding_to):
        self._encoding_to = encoding_to

    def errors(self, errors):
        self._errors = errors

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._encoding_from, self._encoding_to],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        File().convert_encoding(
            fi,
            fo,
            self._encoding_from,
            self._encoding_to,
            self._errors,
        )

        self._logger.info("Encoded file %s" % fi)


class FileArchive(FileBaseTransform):
    """
    Create archeve object.a
    """

    def __init__(self):
        super().__init__()
        self._format = None
        self._create_dir = False

    def format(self, format):
        self._format = format.lower()

    def create_dir(self, create_dir):
        self._create_dir = create_dir

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._src_dir, self._src_pattern, self._format],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)

        if self._dest_dir:
            os.makedirs(self._dest_dir, exist_ok=True)
            dir = self._dest_dir
        else:
            dir = self._src_dir

        valid = EssentialParameters(self.__class__.__name__, [self._dest_name])
        valid()
        dest_path = os.path.join(dir, (self._dest_name + ".%s" % self._format))

        if self._format == "tar":
            with tarfile.open(dest_path, "w") as tar:
                for file in files:
                    arcname = (
                        os.path.join(self._dest_name, os.path.basename(file))
                        if self._create_dir
                        else os.path.basename(file)
                    )
                    tar.add(file, arcname=arcname)
        elif self._format == "zip":
            with zipfile.ZipFile(dest_path, "w") as zp:
                for file in files:
                    arcname = (
                        os.path.join(self._dest_name, os.path.basename(file))
                        if self._create_dir
                        else os.path.basename(file)
                    )
                    zp.write(file, arcname=arcname)
        else:
            raise InvalidParameter("'format' must set one of the followings [tar, zip]")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from requests.auth import HTTPBasicAuth

from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.http import Download


class HttpExtract(BaseStep):
    def __init__(self):
        super().__init__()
        self._src_url = None
        # TODO deprecated.
        self._src_pattern = None
        self._dest_dir = None
        # TODO deprecated.
        self._dest_pattern = None
        self._dest_name = None
        self._timeout = 30
        self._retry_count = 3

    def src_url(self, src_url):
        self._src_url = src_url

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def dest_pattern(self, dest_pattern):
        self._dest_pattern = dest_pattern

    def dest_name(self, dest_name):
        self._dest_name = dest_name

    def timeout(self, timeout):
        self._timeout = timeout

    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def execute(self, *args):
        pass


class HttpDownload(HttpExtract):
    def __init__(self):
        super().__init__()

    def execute(self, *args):
        os.makedirs(self._dest_dir, exist_ok=True)

        if self._src_pattern:
            # Deprecated URL must be a specific string, does not have to be a pattern of regex.
            self._logger.warning(
                "Deprecated URL must be a specific string, does not have to be a pattern of regex."
            )
            valid = EssentialParameters(
                self.__class__.__name__, [self._src_url, self._src_pattern, self._dest_dir]
            )
            valid()
            url = os.path.join(self._src_url, self._src_pattern)
            dest_path = (
                os.path.join(self._dest_dir, self._dest_pattern)
                if self._dest_pattern
                else os.path.join(self._dest_dir, self._src_pattern)
            )
        else:
            valid = EssentialParameters(
                self.__class__.__name__, [self._src_url, self._dest_dir, self._dest_name]
            )
            valid()
            url = self._src_url
            dest_path = os.path.join(self._dest_dir, self._dest_name)

        d = Download(url, dest_path, self._timeout, self._retry_count, **self.get_params())
        d.execute()

    def get_params(self):
        # This parameter will be passed to **kwargs when calling the request.
        # Please implement in subclasses if necessary.
        return {}


class HttpDownloadViaBasicAuth(HttpDownload):
    def __init__(self):
        super().__init__()
        self._user = None
        self._password = None

    def user(self, user):
        self._user = user

    def password(self, password):
        self._password = password

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._user, self._password])
        valid()

        super().execute()

    def get_params(self):
        return {"auth": HTTPBasicAuth(self._user, self._password)}



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import abstractmethod

from cliboa.util.lisboa_log import LisboaLog


class BaseListener(object):
    """
    Base listener for all the listener classes
    """

    def __init__(self):
        self._logger = LisboaLog.get_logger(__name__)


class ScenarioListener(BaseListener):
    """
    Listener for scenario
    """

    @abstractmethod
    def before_scenario(self, worker):
        """
        Execute before scenario start.
        """

    @abstractmethod
    def after_scenario(self, worker):
        """
        Execute after scenario was finished.
        """


class StepListener(BaseListener):
    """
    If you would like to add an extra action for a step,
    create a custom listener class with extend this class,
    and implement any methods below.
    These are called when
    1. before a step is called.
    2. after a step is completed, or when error occured while executing the step.
    3. Very end of the step.
    """

    @abstractmethod
    def before_step(self, *args, **kwargs):
        """
        Execute before a step is called.
        """

    @abstractmethod
    def after_step(self, *args, **kwargs):
        """
        Execute after a step was successfully completed.
        """

    @abstractmethod
    def error_step(self, *args, **kwargs):
        """
        Execute when error occurred while executing a step.
        """

    @abstractmethod
    def after_completion(self, *args, **kwargs):
        """
        Execute after a step
        (no matter the step was successfully completed or ended with an error)
        """


class ScenarioStatusListener(BaseListener):
    """
    Listener for scenario execution status
    """

    def before_scenario(self, worker):
        self._logger.info("Start scenario execution. %s" % (worker.get_scenario_queue_status()))

    def after_scenario(self, worker):
        self._logger.info("Finish scenario execution. %s" % (worker.get_scenario_queue_status()))


class StepStatusListener(StepListener):
    """
    This listener is only for logging.
    By default, Cliboa implements StepStatusListener in all steps.
    """

    def before_step(self, *args, **kwargs):
        self._logger.info("Start step execution. %s" % args[0].__class__.__name__)

    def after_step(self, *args, **kwargs):
        self._logger.info("Finish step execution. %s" % args[0].__class__.__name__)

    def after_completion(self, *args, **kwargs):
        self._logger.info("Complete step execution. %s" % args[0].__class__.__name__)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
from abc import abstractmethod

import jsonlines
import pandas

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.transform.file import FileBaseTransform
from cliboa.util.csv import Csv
from cliboa.util.exception import InvalidParameter


class JsonlToCsvBase(FileBaseTransform):
    """
    Base class of jsonlines transform to csv.
    """

    def __init__(self):
        super().__init__()
        self._quote = "QUOTE_MINIMAL"
        self._after_nl = "LF"
        self._escape_char = None

    def quote(self, quote):
        self._quote = quote

    def after_nl(self, after_nl):
        self._after_nl = after_nl

    def escape_char(self, escape_char):
        self._escape_char = escape_char

    @abstractmethod
    def convert_row(self, row):
        pass

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)
        super().io_files(files, ext="csv", func=self.convert)

    def convert(self, fi, fo):
        writer = None
        with jsonlines.open(fi) as reader, open(
            fo, mode="w", encoding=self._encoding, newline=""
        ) as f:
            for row in reader:
                new_rows = self.convert_row(row)
                if not new_rows:
                    continue
                if not writer:
                    writer = csv.DictWriter(
                        f,
                        new_rows[0].keys(),
                        quoting=Csv.quote_convert(self._quote),
                        lineterminator=Csv.newline_convert(self._after_nl),
                        escapechar=self._escape_char,
                    )
                    writer.writeheader()
                writer.writerows(new_rows)


class JsonlToCsv(JsonlToCsvBase):
    """
    Transform jsonlines to csv.
    """

    def execute(self, *args):
        super().execute()

    def convert_row(self, row):
        return [row]


class JsonlAddKeyValue(FileBaseTransform):
    """
    Insert key value to jsonlines.
    """

    def __init__(self):
        super().__init__()
        self._pairs = {}

    def pairs(self, pairs):
        self._pairs = pairs

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._src_pattern, self._pairs]
        )
        valid()

        if not isinstance(self._pairs, dict):
            raise InvalidParameter("argument 'pairs' only allow dict format.")

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self.check_file_existence(files)
        super().io_files(files, func=self.convert)

    def convert(self, fi, fo):
        with open(fi) as fi, open(fo, mode="w") as fo:
            df = pandas.read_json(fi, orient="records", lines=True)
            for key, value in self._pairs.items():
                df[key] = value
            df.to_json(fo, orient="records", lines=True)



import pymysql

from cliboa.adapter.rdbms import RdbmsSupport


class MysqlAdaptor(RdbmsSupport):
    def get_connection(self, **kwargs):
        # see https://pymysql.readthedocs.io/en/latest/modules/connections.html
        kwargs["host"] = self._host
        kwargs["user"] = self._user
        kwargs["password"] = self._password
        kwargs["db"] = self._dbname
        if self._port:
            kwargs["port"] = self._port
        return pymysql.connect(**kwargs)

    def insert(self, sql, params):
        self._con.cursor().executemany(sql, params)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import os

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.base import BaseStep
from cliboa.util.exception import FileNotFound, InvalidParameter
from cliboa.util.rdbms_util import Rdbms_Util


class BaseRdbms(BaseStep):
    """
    Base class of relational database class.
    """

    def __init__(self):
        super().__init__()

        self._host = None
        self._dbname = None
        self._user = None
        self._password = None
        self._port = None

    def host(self, host):
        self._host = host

    def dbname(self, dbname):
        self._dbname = dbname

    def user(self, user):
        self._user = user

    def password(self, password):
        self._password = password

    def port(self, port):
        self._port = port

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._host, self._dbname, self._user, self._password],
        )
        valid()

    def get_adaptor(self):
        raise NotImplementedError("get_adaptor must be implemented in a subclass")


class BaseRdbmsRead(BaseRdbms):
    def __init__(self):
        super().__init__()
        self._query = None
        self._tblname = None
        self._dest_path = None
        self._encoding = "UTF-8"

    def query(self, query):
        self._query = query

    def tblname(self, tblname):
        self._tblname = tblname

    def dest_path(self, dest_path):
        self._dest_path = dest_path

    def encoding(self, encoding):
        self._encoding = encoding

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._dest_path])
        valid()

        if (self._query and self._tblname) or (not self._query and not self._tblname):
            raise InvalidParameter("Either query or tblname is required.")

        dest_dir = os.path.dirname(self._dest_path)
        if dest_dir:
            os.makedirs(dest_dir, exist_ok=True)

        with self.get_adaptor() as adaptor:
            with open(self._dest_path, mode="w", encoding=self._encoding, newline="") as f:
                if self._tblname:
                    query = Rdbms_Util().select_sql(self._tblname)
                elif isinstance(self._query, str):
                    self._logger.warning(
                        (
                            "DeprecationWarning: "
                            "In the near future, "
                            "the `query` will be changed to accept only dictionary types. "
                        )
                    )
                    query = super()._property_path_reader(self._query)
                else:
                    query_filepath = self._source_path_reader(self._query)
                    with open(query_filepath, "r") as qf:
                        query = qf.read()
                cur = adaptor.select(query)
                writer = None
                for i, row in enumerate(cur):
                    if i == 0:
                        if type(row) is tuple:
                            writer = csv.writer(f, quoting=csv.QUOTE_ALL)
                            columns = [i[0] for i in cur.description]
                            writer.writerow(columns)
                        elif type(row) is dict:
                            writer = csv.DictWriter(f, list(row.keys()), quoting=csv.QUOTE_ALL)
                            writer.writeheader()
                    if writer:
                        writer.writerow(self.callback_handler(row))
                    else:
                        f.write(self.callback_handler(row))

    def callback_handler(self, row):
        if type(row) is tuple:
            return list(row)
        elif type(row) is dict:
            return row
        return row


class BaseRdbmsWrite(BaseRdbms):
    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._tblname = None
        self._encoding = "UTF-8"
        self._chunk_size = 100

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def tblname(self, tblname):
        self._tblname = tblname

    def encoding(self, encoding):
        self._encoding = encoding

    def chunk_size(self, chunk_size):
        self._chunk_size = chunk_size

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._src_pattern, self._tblname]
        )
        valid()

        # Plural files are allowed to insert at the same time,
        # but all files must be the same csv format.
        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            raise FileNotFound("No csv file was found.")
        self._logger.info("Files found %s" % files)

        with open(files[0], mode="r", encoding=self._encoding) as f:
            reader = csv.DictReader(f)
            fieldnames = reader.fieldnames

        rdbmsUtil = Rdbms_Util()
        query = rdbmsUtil.insert_sql(self._tblname, fieldnames)
        self._logger.info("query: %s" % query)
        for file in files:
            with self.get_adaptor() as adaptor:
                tuples = rdbmsUtil.csv_as_params(
                    file, chunk_size=self._chunk_size, encoding=self._encoding
                )
                self._logger.info("tuples: %s" % tuples)
                for params in tuples:
                    self._logger.info("params: %s" % params)
                    adaptor.insert(query, params)



import os

from cryptography.fernet import Fernet

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.base import BaseStep


class AesBase(BaseStep):
    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._dest_dir = None
        self._key_dir = None
        self._key_pattern = None

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def key_dir(self, key_dir):
        self._key_dir = key_dir

    def key_pattern(self, key_pattern):
        self._key_pattern = key_pattern

    def execute(self, *args):
        pass


class AesEncrypt(AesBase):
    def __init__(self):
        super().__init__()

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        if self._dest_dir:
            os.makedirs(self._dest_dir, exist_ok=True)

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            self._logger.info("No files are found. Nothing to do.")
            return

        key_file = super().get_target_files(self._key_dir, self._key_pattern)
        if len(key_file) != 1:
            self._logger.info("Key file must be only one.")
            return

        with open("".join(key_file), "r") as k:
            key = k.read()

        for file in files:
            dest_path = (
                os.path.join(self._dest_dir, os.path.basename(file))
                if self._dest_dir is not None
                else os.path.join(self._src_dir, os.path.basename(file))
            )
            with open(file, "r") as f:
                data = f.read()

            encrypt_data = Fernet(key.encode()).encrypt(data.encode())
            with open(dest_path, "wb") as w:
                w.write(encrypt_data)


class AesDecrypt(AesBase):
    def __init__(self):
        super().__init__()

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        if self._dest_dir:
            os.makedirs(self._dest_dir, exist_ok=True)

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            self._logger.info("No files are found. Nothing to do.")
            return

        key_file = super().get_target_files(self._key_dir, self._key_pattern)
        if len(key_file) != 1:
            self._logger.info("Key file must be only one.")
            return

        with open("".join(key_file), "r") as k:
            key = k.read()

        for file in files:
            dest_path = (
                os.path.join(self._dest_dir, os.path.basename(file))
                if self._dest_dir is not None
                else os.path.join(self._src_dir, os.path.basename(file))
            )
            with open(file, "rb") as f:
                data = f.read()

            decrypt_data = Fernet(key.encode()).decrypt(data)

            with open(dest_path, "wb") as w:
                w.write(decrypt_data)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re

from cliboa.scenario.sftp import BaseSftp
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.cache import ObjectStore
from cliboa.util.constant import StepStatus
from cliboa.util.sftp import Sftp


class SftpExtract(BaseSftp):
    def __init__(self):
        super().__init__()


class SftpDownload(SftpExtract):
    """
    Download files from sftp server
    """

    def __init__(self):
        super().__init__()
        self._quit = False
        self._ignore_empty_file = False

    def quit(self, quit):
        self._quit = quit

    def ignore_empty_file(self, ignore_empty_file):
        self._ignore_empty_file = ignore_empty_file

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._host, self._user, self._src_dir, self._src_pattern],
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        obj = Sftp().list_files(
            dir=self._src_dir,
            dest=self._dest_dir,
            pattern=re.compile(self._src_pattern),
            endfile_suffix=self._endfile_suffix,
            ignore_empty_file=self._ignore_empty_file,
        )

        adaptor = super().get_adaptor()
        files = adaptor.execute(obj)

        if self._quit is True and len(files) == 0:
            self._logger.info("No file was found. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        self._logger.info("Files downloaded %s" % files)

        # cache downloaded file names
        ObjectStore.put(self._step, files)


class SftpDelete(SftpExtract):
    """
    Delete file from sftp server
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._host, self._user, self._src_dir, self._src_pattern],
        )
        valid()

        obj = Sftp().clear_files(
            dir=self._src_dir,
            pattern=re.compile(self._src_pattern),
        )

        adaptor = super().get_adaptor()
        adaptor.execute(obj)


class SftpDownloadFileDelete(SftpExtract):
    """
    Delete all downloaded files.
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        files = ObjectStore.get(self._symbol)

        if files is not None and len(files) > 0:
            self._logger.info("Delete files %s" % files)

            self._host = super().get_step_argument("host")
            self._user = super().get_step_argument("user")
            self._password = super().get_step_argument("password")
            self._key = super().get_step_argument("key")
            self._timeout = super().get_step_argument("timeout")
            self._retry_count = super().get_step_argument("retry_count")
            self._port = super().get_step_argument("port")
            self._endfile_suffix = super().get_step_argument("endfile_suffix")
            self._src_dir = super().get_step_argument("src_dir")

            adaptor = super().get_adaptor()

            endfile_suffix = super().get_step_argument("endfile_suffix")
            for file in files:
                obj = Sftp().remove_specific_file(
                    dir=self._src_dir,
                    fname=file,
                )
                adaptor.execute(obj)
                self._logger.info("%s is successfully deleted." % file)

                if endfile_suffix:
                    obj = Sftp().remove_specific_file(
                        dir=self._src_dir,
                        fname=file + endfile_suffix,
                    )
                    self._logger.info("%s is successfully deleted." % (file + endfile_suffix))
        else:
            self._logger.info("No files to delete.")


class SftpFileExistsCheck(SftpExtract):
    """
    File check in sftp server
    """

    def __init__(self):
        super().__init__()
        self._ignore_empty_file = False

    def ignore_empty_file(self, ignore_empty_file):
        self._ignore_empty_file = ignore_empty_file

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._host, self._user, self._src_dir, self._src_pattern],
        )
        valid()

        obj = Sftp().file_exists_check(
            dir=self._src_dir,
            pattern=re.compile(self._src_pattern),
            ignore_empty_file=self._ignore_empty_file,
        )

        adaptor = super().get_adaptor()
        files = adaptor.execute(obj)

        if len(files) == 0:
            self._logger.info("File not found. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        self._logger.info("File was found. After process will be processed")
